import importlib
import pkgutil
__all__ = [] # expose objax implementation as base nn
module = importlib.import_module('.'+'objax',package=__name__)
globals().update({k: getattr(module, k) for k in module.__all__})
__all__ += module.__all__



# import importlib
# import pkgutil
# __all__ = []
# for loader, module_name, is_pkg in  pkgutil.walk_packages(__path__):
#     module = importlib.import_module('.'+module_name,package=__name__)
#     try: 
#         globals().update({k: getattr(module, k) for k in module.__all__})
#         __all__ += module.__all__
#     except AttributeError: continue
# # concatenate the __all__ from each of the submodules (expose to user)

__version__ = '1.0.1'
from .nn import *
from .groups import *
from .reps import *



import jax
import jax.numpy as jnp
import objax.nn as nn
import objax.functional as F
import numpy as np
from emlp.reps import T,Rep,Scalar
from emlp.reps import bilinear_weights
from emlp.reps.product_sum_reps import SumRep
import collections
from emlp.utils import Named,export
import scipy as sp
import scipy.special
import random
import logging
from objax.variable import TrainVar, StateVar
from objax.nn.init import kaiming_normal, xavier_normal
from objax.module import Module
import objax
from objax.nn.init import orthogonal
from scipy.special import binom
from jax import jit,vmap
from functools import lru_cache as cache

def Sequential(*args):
    """ Wrapped to mimic pytorch syntax"""
    return nn.Sequential(args)

@export
class Linear(nn.Linear):
    """ Basic equivariant Linear layer from repin to repout."""
    def __init__(self, repin, repout):
        nin,nout = repin.size(),repout.size()
        super().__init__(nin,nout)
        self.b = TrainVar(objax.random.uniform((nout,))/jnp.sqrt(nout))
        self.w = TrainVar(orthogonal((nout, nin)))
        self.rep_W = rep_W = repout*repin.T
        
        rep_bias = repout
        self.Pw = rep_W.equivariant_projector()
        self.Pb = rep_bias.equivariant_projector()
        logging.info(f"Linear W components:{rep_W.size()} rep:{rep_W}")
    def __call__(self, x): # (cin) -> (cout)
        logging.debug(f"linear in shape: {x.shape}")
        W = (self.Pw@self.w.value.reshape(-1)).reshape(*self.w.value.shape)
        b = self.Pb@self.b.value
        out = x@W.T+b
        logging.debug(f"linear out shape:{out.shape}")
        return out

@export
class BiLinear(Module):
    """ Cheap bilinear layer (adds parameters for each part of the input which can be
        interpreted as a linear map from a part of the input to the output representation)."""
    def __init__(self, repin, repout):
        super().__init__()
        Wdim, weight_proj = bilinear_weights(repout,repin)
        self.weight_proj = jit(weight_proj)
        self.w = TrainVar(objax.random.normal((Wdim,)))
        logging.info(f"BiW components: dim:{Wdim}")

    def __call__(self, x,training=True):
        # compatible with non sumreps? need to check
        W = self.weight_proj(self.w.value,x)
        out= .1*(W@x[...,None])[...,0]
        return out

@export
def gated(ch_rep:Rep) -> Rep:
    """ Returns the rep with an additional scalar 'gate' for each of the nonscalars and non regular
        reps in the input. To be used as the output for linear (and or bilinear) layers directly
        before a :func:`GatedNonlinearity` to produce its scalar gates. """
    if isinstance(ch_rep,SumRep):
        return ch_rep+sum([Scalar(rep.G) for rep in ch_rep if rep!=Scalar and not rep.is_permutation])
    else:
        return ch_rep+Scalar(ch_rep.G) if not ch_rep.is_permutation else ch_rep

@export
class GatedNonlinearity(Module):
    """ Gated nonlinearity. Requires input to have the additional gate scalars
        for every non regular and non scalar rep. Applies swish to regular and
        scalar reps. """
    def __init__(self,rep):
        super().__init__()
        self.rep=rep
    def __call__(self,values):
        gate_scalars = values[..., gate_indices(self.rep)]
        activations = jax.nn.sigmoid(gate_scalars) * values[..., :self.rep.size()]
        return activations

@export
class EMLPBlock(Module):
    """ Basic building block of EMLP consisting of G-Linear, biLinear,
        and gated nonlinearity. """
    def __init__(self,rep_in,rep_out):
        super().__init__()
        self.linear = Linear(rep_in,gated(rep_out))
        self.bilinear = BiLinear(gated(rep_out),gated(rep_out))
        self.nonlinearity = GatedNonlinearity(rep_out)
    def __call__(self,x):
        lin = self.linear(x)
        preact =self.bilinear(lin)+lin
        return self.nonlinearity(preact)

def uniform_rep_general(ch,*rep_types):
    """ adds all combinations of (powers of) rep_types up to
        a total size of ch channels. """
    raise NotImplementedError
    
    
    

@export
def uniform_rep(ch,group):
    """ A heuristic method for allocating a given number of channels (ch)
        into tensor types. Attempts to distribute the channels evenly across
        the different tensor types. Useful for hands off layer construction.
        
        Args:
            ch (int): total number of channels
            group (Group): symmetry group

        Returns:
            SumRep: The direct sum representation with dim(V)=ch
        """
    d = group.d
    Ns = np.zeros((lambertW(ch,d)+1,),int) # number of tensors of each rank
    while ch>0:
        max_rank = lambertW(ch,d) # compute the max rank tensor that can fit up to
        Ns[:max_rank+1] += np.array([d**(max_rank-r) for r in range(max_rank+1)],dtype=int)
        ch -= (max_rank+1)*d**max_rank # compute leftover channels
    sum_rep = sum([binomial_allocation(nr,r,group) for r,nr in enumerate(Ns)])
    sum_rep,perm = sum_rep.canonicalize()
    return sum_rep

def lambertW(ch,d):
    """ Returns solution to x*d^x = ch rounded down."""
    max_rank=0
    while (max_rank+1)*d**max_rank <= ch:
        max_rank += 1
    max_rank -= 1
    return max_rank

def binomial_allocation(N,rank,G):
    """ Allocates N of tensors of total rank r=(p+q) into
        T(k,r-k) for k=0,1,...,r to match the binomial distribution.
        For orthogonal representations there is no
        distinction between p and q, so this op is equivalent to N*T(rank)."""
    if N==0: return 0
    n_binoms = N//(2**rank)
    n_leftover = N%(2**rank)
    even_split = sum([n_binoms*int(binom(rank,k))*T(k,rank-k,G) for k in range(rank+1)])
    ps = np.random.binomial(rank,.5,n_leftover)
    ragged = sum([T(int(p),rank-int(p),G) for p in ps])
    out = even_split+ragged
    return out

def uniform_allocation(N,rank):
    """ Uniformly allocates N of tensors of total rank r=(p+q) into
        T(k,r-k) for k=0,1,...,r. For orthogonal representations there is no
        distinction between p and q, so this op is equivalent to N*T(rank)."""
    if N==0: return 0
    even_split = sum((N//(rank+1))*T(k,rank-k) for k in range(rank+1))
    ragged = sum(random.sample([T(k,rank-k) for k in range(rank+1)],N%(rank+1)))
    return even_split+ragged

@export
class EMLP(Module,metaclass=Named):
    """ Equivariant MultiLayer Perceptron. 
        If the input ch argument is an int, uses the hands off uniform_rep heuristic.
        If the ch argument is a representation, uses this representation for the hidden layers.
        Individual layer representations can be set explicitly by using a list of ints or a list of
        representations, rather than use the same for each hidden layer.

        Args:
            rep_in (Rep): input representation
            rep_out (Rep): output representation
            group (Group): symmetry group
            ch (int or list[int] or Rep or list[Rep]): number of channels in the hidden layers
            num_layers (int): number of hidden layers

        Returns:
            Module: the EMLP objax module."""
    def __init__(self,rep_in,rep_out,group,ch=384,num_layers=3):#@
        super().__init__()
        logging.info("Initing EMLP (objax)")
        self.rep_in =rep_in(group)
        self.rep_out = rep_out(group)
        
        self.G=group
        # Parse ch as a single int, a sequence of ints, a single Rep, a sequence of Reps
        if isinstance(ch,int): middle_layers = num_layers*[uniform_rep(ch,group)]#[uniform_rep(ch,group) for _ in range(num_layers)]
        elif isinstance(ch,Rep): middle_layers = num_layers*[ch(group)]
        else: middle_layers = [(c(group) if isinstance(c,Rep) else uniform_rep(c,group)) for c in ch]
        #assert all((not rep.G is None) for rep in middle_layers[0].reps)
        reps = [self.rep_in]+middle_layers
        logging.info(f"Reps: {reps}")
        self.network = Sequential(
            *[EMLPBlock(rin,rout) for rin,rout in zip(reps,reps[1:])],
            Linear(reps[-1],self.rep_out)
        )
    def __call__(self,x,training=True):
        return self.network(x)

def swish(x):
    return jax.nn.sigmoid(x)*x

def MLPBlock(cin,cout):
    return Sequential(nn.Linear(cin,cout),swish)#,nn.BatchNorm0D(cout,momentum=.9),swish)#,

@export
class MLP(Module,metaclass=Named):
    """ Standard baseline MLP. Representations and group are used for shapes only. """
    def __init__(self,rep_in,rep_out,group,ch=384,num_layers=3):
        super().__init__()
        self.rep_in =rep_in(group)
        self.rep_out = rep_out(group)
        self.G = group
        chs = [self.rep_in.size()] + num_layers*[ch]
        cout = self.rep_out.size()
        logging.info("Initing MLP")
        self.net = Sequential(
            *[MLPBlock(cin,cout) for cin,cout in zip(chs,chs[1:])],
            nn.Linear(chs[-1],cout)
        )
    def __call__(self,x,training=True):
        y = self.net(x)
        return y

@export
class Standardize(Module):
    """ A convenience module to wrap a given module, normalize its input
        by some dataset x mean and std stats, and unnormalize its output by
        the dataset y mean and std stats. 

        Args:
            model (Module): model to wrap
            ds_stats ((μx,σx,μy,σy) or (μx,σx)): tuple of the normalization stats
        
        Returns:
            Module: Wrapped model with input normalization (and output unnormalization)"""
    def __init__(self,model,ds_stats):
        super().__init__()
        self.model = model
        self.ds_stats=ds_stats
    def __call__(self,x,training):
        if len(self.ds_stats)==2:
            muin,sin = self.ds_stats
            return self.model((x-muin)/sin,training=training)
        else:
            muin,sin,muout,sout = self.ds_stats
            y = sout*self.model((x-muin)/sin,training=training)+muout
            return y



# Networks for hamiltonian dynamics (need to sum for batched Hamiltonian grads)
@export
class MLPode(Module,metaclass=Named):
    def __init__(self,rep_in,rep_out,group,ch=384,num_layers=3):
        super().__init__()
        self.rep_in =rep_in(group)
        self.rep_out = rep_out(group)
        self.G = group
        chs = [self.rep_in.size()] + num_layers*[ch]
        cout = self.rep_out.size()
        logging.info("Initing MLP")
        self.net = Sequential(
            *[Sequential(nn.Linear(cin,cout),swish) for cin,cout in zip(chs,chs[1:])],
            nn.Linear(chs[-1],cout)
        )
    def __call__(self,z,t):
        return self.net(z)

@export
class EMLPode(EMLP):
    """ Neural ODE Equivariant MLP. Same args as EMLP."""
    #__doc__ += EMLP.__doc__.split('.')[1]
    def __init__(self,rep_in,rep_out,group,ch=384,num_layers=3):#@
        #super().__init__()
        logging.info("Initing EMLP")
        self.rep_in =rep_in(group)
        self.rep_out = rep_out(group)
        self.G=group
        # Parse ch as a single int, a sequence of ints, a single Rep, a sequence of Reps
        if isinstance(ch,int): middle_layers = num_layers*[uniform_rep(ch,group)]#[uniform_rep(ch,group) for _ in range(num_layers)]
        elif isinstance(ch,Rep): middle_layers = num_layers*[ch(group)]
        else: middle_layers = [(c(group) if isinstance(c,Rep) else uniform_rep(c,group)) for c in ch]
        #print(middle_layers[0].reps[0].G)
        #print(self.rep_in.G)
        reps = [self.rep_in]+middle_layers
        logging.info(f"Reps: {reps}")
        self.network = Sequential(
            *[EMLPBlock(rin,rout) for rin,rout in zip(reps,reps[1:])],
            Linear(reps[-1],self.rep_out)
        )
    def __call__(self,z,t):
        return self.network(z)

# Networks for hamiltonian dynamics (need to sum for batched Hamiltonian grads)
@export
class MLPH(Module,metaclass=Named):
    def __init__(self,rep_in,rep_out,group,ch=384,num_layers=3):
        super().__init__()
        self.rep_in =rep_in(group)
        self.rep_out = rep_out(group)
        self.G = group
        chs = [self.rep_in.size()] + num_layers*[ch]
        cout = self.rep_out.size()
        logging.info("Initing MLP")
        self.net = Sequential(
            *[Sequential(nn.Linear(cin,cout),swish) for cin,cout in zip(chs,chs[1:])],
            nn.Linear(chs[-1],cout)
        )
    def H(self,x):#,training=True):
        y = self.net(x).sum()
        return y
    def __call__(self,x):
        return self.H(x)

@export
class EMLPH(EMLP):
    """ Equivariant EMLP modeling a Hamiltonian for HNN. Same args as EMLP"""
    #__doc__ += EMLP.__doc__.split('.')[1]
    def H(self,x):#,training=True):
        y = self.network(x)
        return y.sum()
    def __call__(self,x):
        return self.H(x)

@export
@cache(maxsize=None)
def gate_indices(ch_rep:Rep) -> jnp.ndarray:
    """ Indices for scalars, and also additional scalar gates
        added by gated(sumrep)"""
    channels = ch_rep.size()
    perm = ch_rep.perm
    indices = np.arange(channels)

    if not isinstance(ch_rep,SumRep): # If just a single rep, only one scalar at end
        return indices if ch_rep.is_permutation else np.ones(ch_rep.size())*ch_rep.size()

    num_nonscalars = 0
    i=0
    for rep in ch_rep:
        if rep!=Scalar and not rep.is_permutation:
            indices[perm[i:i+rep.size()]] = channels+num_nonscalars
            num_nonscalars+=1
        i+=rep.size()
    return indices


import sys

class Named(type):
    def __str__(self):
        return self.__name__
    def __repr__(self):
        return self.__name__

def export(fn):
    mod = sys.modules[fn.__module__]
    if hasattr(mod, '__all__'):
        mod.__all__.append(fn.__name__)
    else:
        mod.__all__ = [fn.__name__]
    return fn


import numpy as np
import jax.numpy as jnp
from emlp.reps import Scalar,Vector,T
from emlp.utils import export,Named
from emlp.groups import SO,O,Trivial,Lorentz,RubiksCube,Cube
from functools import partial
import itertools
from jax import vmap,jit
from objax import Module

@export
class Inertia(object):
    def __init__(self,N=1024,k=5):
        super().__init__()
        self.dim = (1+3)*k
        self.X = np.random.randn(N,self.dim)
        self.X[:,:k] = np.log(1+np.exp(self.X[:,:k])) # Masses
        mi = self.X[:,:k]
        ri = self.X[:,k:].reshape(-1,k,3)
        I = np.eye(3)
        r2 = (ri**2).sum(-1)[...,None,None]
        inertia = (mi[:,:,None,None]*(r2*I - ri[...,None]*ri[...,None,:])).sum(1)
        self.Y = inertia.reshape(-1,9)
        self.rep_in = k*Scalar+k*Vector
        self.rep_out = T(2)
        self.symmetry = O(3)
        # One has to be careful computing offset and scale in a way so that standardizing
        # does not violate equivariance
        Xmean = self.X.mean(0)
        Xmean[k:] = 0
        Xstd = np.zeros_like(Xmean)
        Xstd[:k] = np.abs(self.X[:,:k]).mean(0)#.std(0)
        #Xstd[k:] = (np.sqrt((self.X[:,k:].reshape(N,k,3)**2).mean((0,2))[:,None]) + np.zeros((k,3))).reshape(k*3)
        Xstd[k:] = (np.abs(self.X[:,k:].reshape(N,k,3)).mean((0,2))[:,None] + np.zeros((k,3))).reshape(k*3)
        Ymean = 0*self.Y.mean(0)
        #Ystd = np.sqrt(((self.Y-Ymean)**2).mean((0,1)))+ np.zeros_like(Ymean)
        Ystd = np.abs(self.Y-Ymean).mean((0,1)) + np.zeros_like(Ymean)
        self.stats =0,1,0,1#Xmean,Xstd,Ymean,Ystd

    def __getitem__(self,i):
        return (self.X[i],self.Y[i])
    def __len__(self):
        return self.X.shape[0]
    def default_aug(self,model):
        return GroupAugmentation(model,self.rep_in,self.rep_out,self.symmetry)

@export
class O5Synthetic(object):
    def __init__(self,N=1024):
        super().__init__()
        d=5
        self.dim = 2*d
        self.X = np.random.randn(N,self.dim)
        ri = self.X.reshape(-1,2,5)
        r1,r2 = ri.transpose(1,0,2)
        self.Y = np.sin(np.sqrt((r1**2).sum(-1)))-.5*np.sqrt((r2**2).sum(-1))**3 + (r1*r2).sum(-1)/(np.sqrt((r1**2).sum(-1))*np.sqrt((r2**2).sum(-1)))
        self.rep_in = 2*Vector
        self.rep_out = Scalar
        self.symmetry = O(d)
        self.Y = self.Y[...,None]
        # One has to be careful computing mean and std in a way so that standardizing
        # does not violate equivariance
        Xmean = self.X.mean(0) # can add and subtract arbitrary tensors
        Xscale = (np.sqrt((self.X.reshape(N,2,d)**2).mean((0,2)))[:,None]+0*ri[0]).reshape(self.dim)
        self.stats = 0,Xscale,self.Y.mean(axis=0),self.Y.std(axis=0)

    def __getitem__(self,i):
        return (self.X[i],self.Y[i])
    def __len__(self):
        return self.X.shape[0]
    def default_aug(self,model):
        return GroupAugmentation(model,self.rep_in,self.rep_out,self.symmetry)

@export
class ParticleInteraction(object):
    """ Electron muon e^4 interaction"""
    def __init__(self,N=1024):
        super().__init__()
        self.dim = 4*4
        self.rep_in = 4*Vector
        self.rep_out = Scalar
        self.X = np.random.randn(N,self.dim)/4
        P = self.X.reshape(N,4,4)
        p1,p2,p3,p4 = P.transpose(1,0,2)
        𝜂 = np.diag(np.array([1.,-1.,-1.,-1.]))
        dot = lambda v1,v2: ((v1@𝜂)*v2).sum(-1)
        Le = (p1[:,:,None]*p3[:,None,:] - (dot(p1,p3)-dot(p1,p1))[:,None,None]*𝜂)
        L𝜇 = ((p2@𝜂)[:,:,None]*(p4@𝜂)[:,None,:] - (dot(p2,p4)-dot(p2,p2))[:,None,None]*𝜂)
        M = 4*(Le*L𝜇).sum(-1).sum(-1)
        self.Y = M
        self.symmetry = Lorentz()
        self.Y = self.Y[...,None]
        # One has to be careful computing mean and std in a way so that standardizing
        # does not violate equivariance
        self.Xscale = np.sqrt((np.abs((self.X.reshape(N,4,4)@𝜂)*self.X.reshape(N,4,4)).mean(-1)).mean(0))
        self.Xscale = (self.Xscale[:,None]+np.zeros((4,4))).reshape(-1)
        self.stats = 0,self.Xscale,self.Y.mean(axis=0),self.Y.std(axis=0)#self.X.mean(axis=0),self.X.std(axis=0),
    def __getitem__(self,i):
        return (self.X[i],self.Y[i])
    def __len__(self):
        return self.X.shape[0]
    def default_aug(self,model):
        return GroupAugmentation(model,self.rep_in,self.rep_out,self.symmetry)

class GroupAugmentation(Module):
    def __init__(self,network,rep_in,rep_out,group):
        super().__init__()
        self.rep_in=rep_in
        self.rep_out=rep_out
        self.G=group
        self.rho_in = jit(vmap(self.rep_in.rho))
        self.rho_out = jit(vmap(self.rep_out.rho))
        self.model = network
    def __call__(self,x,training=True):
        if training:
            gs = self.G.samples(x.shape[0])
            rhout_inv = jnp.linalg.inv(self.rho_out(gs))
            return (rhout_inv@self.model((self.rho_in(gs)@x[...,None])[...,0],training)[...,None])[...,0]
        else:
            return self.model(x,False)

@export
class InvertedCube(object):
    def __init__(self,train=True):
        pass #TODO: finish implementing this simple dataset
        solved_state = np.eye(6)
        parity_perm = np.array([5,3,4,1,2,0])
        parity_state = solved_state[:,parity_perm]

        labels = np.array([1,0]).astype(int)
        self.X = np.zeros((2,6*6))
        self.X[0] = solved_state.reshape(-1)
        self.X[1] = parity_state.reshape(-1)
        self.Y = labels
        self.symmetry = Cube()
        self.rep_in = 6*Vector
        self.rep_out = 2*Scalar
        self.stats = (0,1)
        if train==False: # Scramble the cubes for test time
            gs = self.symmetry.samples(100)
            self.X = np.repeat(self.X,50,axis=0).reshape(100,6,6)@gs
            self.Y = np.repeat(self.Y,50,axis=0)
            p = np.random.permutation(100)
            self.X = self.X[p].reshape((100,-1))
            self.Y = self.Y[p].reshape((100,))
        self.X = np.array(self.X)
        self.Y = np.array(self.Y)
            
    def __getitem__(self,i):
        return (self.X[i],self.Y[i])
    def __len__(self):
        return self.X.shape[0]


#### Ways of constructing invalid Rubik's cubes 
# (see https://ruwix.com/the-rubiks-cube/unsolvable-rubiks-cube-invalid-scramble/)

def UBedge_flip(state):
    """ # invalid edge flip on Rubiks state (6,48)"""
    UB_edge = np.array([1,3*8+1]) # Top middle of U, top middle of B
    edge_flip = state.copy()
    edge_flip[:,UB_edge] = edge_flip[:,np.roll(UB_edge,1)]
    return edge_flip

def ULBcorner_rot(state,i=1):
    """ Invalid rotated corner with ULB corner rotation """
    ULB_corner_ids = np.array([0,4*8,3*8+2]) # top left of U, top left of L, top right of B
    rotated_corner_state = state.copy()
    rotated_corner_state[:,ULB_corner_ids] = rotated_corner_state[:,np.roll(ULB_corner_ids,i)]
    return rotated_corner_state

def LBface_swap(state):
    """ Invalid piece swap between L center top and B center top """
    L_B_center_top_faces = np.array([4*8+1,3*8+1])
    piece_swap = state.copy()
    piece_swap[:,L_B_center_top_faces] = piece_swap[:,np.roll(L_B_center_top_faces,1)]
    return piece_swap


@export
class BrokenRubiksCube(object):
    """ Binary classification problem of predicting whether a Rubik's cube configuration
        is solvable or 'broken' and not able to be solved by transformations from the group
        e.g. by removing and twisting a corner before replacing.
        Dataset is generated by taking several hand identified simple instances of solvable
        and unsolvable cubes, and then scrambling them.
        
        Features are represented as 6xT(1) tensors of the Rubiks Group (one hot for each color)"""
    def __init__(self,train=True):
        super().__init__()
        # start with a valid configuration
        
        solved_state = np.zeros((6,48))
        for i in range(6):
            solved_state[i,8*i:8*(i+1)] = 1

        Id = lambda x: x
        transforms =  [Id,itertools.product([Id,ULBcorner_rot,partial(ULBcorner_rot,i=2)],
                                        [Id,UBedge_flip],
                                        [Id,LBface_swap])]
        #equivalence_classes = np.vstack([t3(t2(t1(solved_state))) for t1,t2,t3 in transforms])
        labels = np.zeros((22,))
        labels[:11]=1 # First configuration is solvable
        labels[11:]=0 # all others are not
        self.X = np.zeros((22,6*48)) # duplicate solvable example 11 times for convencience (balance)
        self.X[:11] = solved_state.reshape(-1)#equivalence_classes.reshape(12,-1)[:1]
        parity_perm = np.array([5,3,4,1,2,0])
        self.X[11:] = solved_state.reshape(6,6,8)[:,parity_perm].reshape(-1)#equivalence_classes.reshape(12,-1)[1:]
        self.Y = labels
        self.symmetry = RubiksCube()
        self.rep_in = 6*Vector
        self.rep_out = 2*Scalar
        self.stats = (0,1)
        if train==False: # Scramble the cubes for test time
            gs = self.symmetry.samples(440)
            self.X = np.repeat(self.X,20,axis=0).reshape(440,6,48)@gs
            self.Y = np.repeat(self.Y,20,axis=0)
            p = np.random.permutation(440)
            self.X = self.X[p].reshape((440,-1))
            self.Y = self.Y[p].reshape((440,))
        self.X = np.array(self.X)
        self.Y = np.array(self.Y)
            
    def __getitem__(self,i):
        return (self.X[i],self.Y[i])
    def __len__(self):
        return self.X.shape[0]

# @export
# class BrokenRubiksCube2x2(object):
#     """ Binary classification problem of predicting whether a Rubik's cube configuration
#         is solvable or 'broken' and not able to be solved by transformations from the group
#         e.g. by removing and twisting a corner before replacing.
#         Dataset is generated by taking several hand identified simple instances of solvable
#         and unsolvable cubes, and then scrambling them.
        
#         Features are represented as 6xT(1) tensors of the Rubiks Group (one hot for each color)"""
#     def __init__(self,train=True):
#         super().__init__()
#         # start with a valid configuration
        
#         solved_state = np.zeros((6,24))
#         for i in range(6):
#             solved_state[i,4*i:4*(i+1)] = 1

#         ULB_corner_ids = np.array([0,4*4,3*4+2]) # top left of U, top left of L, top right of B
#         rotated_corner_state = solved_state.copy()
#         rotated_corner_state[:,ULB_corner_ids] = rotated_corner_state[:,np.roll(ULB_corner_ids,1)]
        
#         labels = np.zeros((2,))
#         labels[:1]=1 # First configuration is solvable
#         labels[1:]=0 # all others are not
#         self.X = np.zeros((2,6*24)) # duplicate solvable example 11 times for convencience (balance)
#         self.X[0] = solved_state.reshape(-1)
#         parity_perm = np.array([5,3,4,1,2,0])
#         self.X[1] = rotated_corner_state.reshape(6,6,4)[:,parity_perm].reshape(-1)
#         self.Y = labels
#         self.X = np.repeat(self.X,10,axis=0)
#         self.Y = np.repeat(self.Y,10,axis=0)
#         self.symmetry = RubiksCube2x2()
#         self.rep_in = 6*Vector
#         self.rep_out = 2*Scalar
#         self.stats = (0,1)
#         if train==False: # Scramble the cubes for test time
#             N = 200
#             gs = self.symmetry.samples(N)
#             self.X = np.repeat(self.X,10,axis=0).reshape(N,6,24)@gs
#             self.Y = np.repeat(self.Y,10,axis=0)
#             p = np.random.permutation(N)
#             self.X = self.X[p].reshape((N,-1))
#             self.Y = self.Y[p].reshape((N,))
#         self.X = np.array(self.X)
#         self.Y = np.array(self.Y)
            
#     def __getitem__(self,i):
#         return (self.X[i],self.Y[i])
#     def __len__(self):
#         return self.X.shape[0]


import jax
import jax.numpy as jnp
import numpy as np
from emlp.reps import T,Rep,Scalar
from emlp.reps import bilinear_weights
#from emlp.reps import LinearOperator # why does this not work?
from emlp.reps.linear_operator_base import LinearOperator
from emlp.reps.product_sum_reps import SumRep
from emlp.groups import Group
from emlp.utils import Named,export
from flax import linen as nn
import logging
from emlp.nn import gated,gate_indices,uniform_rep
from typing import Union,Iterable,Optional
# def Sequential(*args):
#     """ Wrapped to mimic pytorch syntax"""
#     return nn.Sequential(args)




@export
def Linear(repin,repout):
    """ Basic equivariant Linear layer from repin to repout."""
    cin =repin.size()
    cout = repout.size()
    rep_W = repin>>repout
    Pw = rep_W.equivariant_projector()
    Pb = repout.equivariant_projector()
    logging.info(f"Linear W components:{rep_W.size()} rep:{rep_W}")
    return _Linear(Pw,Pb,cout)

class _Linear(nn.Module):
    Pw:LinearOperator
    Pb:LinearOperator
    cout:int
    @nn.compact
    def __call__(self,x):
        w = self.param('w',nn.initializers.lecun_normal(),(self.cout,x.shape[-1]))
        b = self.param('b',nn.initializers.zeros,(self.cout,))
        W = (self.Pw@w.reshape(-1)).reshape(*w.shape)
        B = self.Pb@b
        return x@W.T+B

@export
def BiLinear(repin,repout):
    """ Cheap bilinear layer (adds parameters for each part of the input which can be
        interpreted as a linear map from a part of the input to the output representation)."""
    Wdim, weight_proj = bilinear_weights(repout,repin)
        #self.w = TrainVar(objax.random.normal((Wdim,)))#xavier_normal((Wdim,))) #TODO: revert to xavier
    logging.info(f"BiW components: dim:{Wdim}")
    return _BiLinear(Wdim,weight_proj)

class _BiLinear(nn.Module):
    Wdim:int
    weight_proj:callable

    @nn.compact
    def __call__(self, x):
        w = self.param('w',nn.initializers.normal(),(self.Wdim,)) #TODO: change to standard normal
        W = self.weight_proj(w,x)
        out= .1*(W@x[...,None])[...,0]
        return out


@export
class GatedNonlinearity(nn.Module): #TODO: add support for mixed tensors and non sumreps
    """ Gated nonlinearity. Requires input to have the additional gate scalars
        for every non regular and non scalar rep. Applies swish to regular and
        scalar reps. (Right now assumes rep is a SumRep)"""
    rep:Rep
    def __call__(self,values):
        gate_scalars = values[..., gate_indices(self.rep)]
        activations = jax.nn.sigmoid(gate_scalars) * values[..., :self.rep.size()]
        return activations

@export
def EMLPBlock(rep_in,rep_out):
    """ Basic building block of EMLP consisting of G-Linear, biLinear,
        and gated nonlinearity. """
    linear = Linear(rep_in,gated(rep_out))
    bilinear = BiLinear(gated(rep_out),gated(rep_out))
    nonlinearity = GatedNonlinearity(rep_out)
    return _EMLPBlock(linear,bilinear,nonlinearity)

class _EMLPBlock(nn.Module):
    linear:nn.Module
    bilinear:nn.Module
    nonlinearity:nn.Module

    def __call__(self,x):
        lin = self.linear(x)
        preact =self.bilinear(lin)+lin
        return self.nonlinearity(preact)

@export
def EMLP(rep_in,rep_out,group,ch=384,num_layers=3):
    """ Equivariant MultiLayer Perceptron. 
        If the input ch argument is an int, uses the hands off uniform_rep heuristic.
        If the ch argument is a representation, uses this representation for the hidden layers.
        Individual layer representations can be set explicitly by using a list of ints or a list of
        representations, rather than use the same for each hidden layer.

        Args:
            rep_in (Rep): input representation
            rep_out (Rep): output representation
            group (Group): symmetry group
            ch (int or list[int] or Rep or list[Rep]): number of channels in the hidden layers
            num_layers (int): number of hidden layers

        Returns:
            Module: the EMLP objax module."""
    logging.info("Initing EMLP (flax)")
    rep_in = rep_in(group)
    rep_out = rep_out(group)
    if isinstance(ch,int): middle_layers = num_layers*[uniform_rep(ch,group)]
    elif isinstance(ch,Rep): middle_layers = num_layers*[ch(group)]
    else: middle_layers = [(c(group) if isinstance(c,Rep) else uniform_rep(c,group)) for c in ch]
    reps = [rep_in]+middle_layers
    logging.info(f"Reps: {reps}")
    return Sequential(*[EMLPBlock(rin,rout) for rin,rout in zip(reps,reps[1:])],Linear(reps[-1],rep_out))

def swish(x):
    return jax.nn.sigmoid(x)*x

class _Sequential(nn.Module):
    modules:Iterable[callable]
    def __call__(self,x):
        for module in self.modules:
            x = module(x)
        return x

def Sequential(*layers):
    return _Sequential(layers)

def MLPBlock(cout):
    return Sequential(nn.Dense(cout),swish)  # ,nn.BatchNorm0D(cout,momentum=.9),swish)#,

@export
class MLP(nn.Module,metaclass=Named):
    """ Standard baseline MLP. Representations and group are used for shapes only. """
    rep_in:Rep
    rep_out:Rep
    group:Group
    ch: Optional[InterruptedError]=384
    num_layers:Optional[int]=3
    def setup(self):
        logging.info("Initing MLP (flax)")
        cout = self.rep_out(self.group).size()
        self.modules = [MLPBlock(self.ch) for _ in range(self.num_layers)]+[nn.Dense(cout)]
    def __call__(self,x):
        for module in self.modules:
            x = module(x)
        return x



# from .product_sum_reps import SumRep,DeferredSumRep,ProductRep,DeferredProductRep,DirectProduct
# __all__=["SumRep","DeferredSumRep","ProductRep","DeferredProductRep","DirectProduct"]
import importlib
import pkgutil
__all__ = []
for loader, module_name, is_pkg in pkgutil.walk_packages(__path__):
    module = importlib.import_module('.'+module_name,package=__name__)
    try: 
        globals().update({k: getattr(module, k) for k in module.__all__})
        __all__ += module.__all__
    except AttributeError: continue

# concatenate __all__ from each of the modules


import jax
import jax.numpy as jnp
import numpy as np
from emlp.reps import Rep
from emlp.reps import bilinear_weights
from emlp.utils import export
import logging
import haiku as hk
from emlp.nn import gated,gate_indices,uniform_rep

def Sequential(*args):
    """ Wrapped to mimic pytorch syntax"""
    return lambda x: hk.Sequential(args)(x)

@export
def Linear(repin,repout):
    rep_W = repout << repin
    logging.info(f"Linear W components:{rep_W.size()} rep:{rep_W}")
    rep_bias = repout
    Pw = rep_W.equivariant_projector()
    Pb = rep_bias.equivariant_projector()
    return lambda x: hkLinear(Pw,Pb,(repout.size(),repin.size()))(x)

class hkLinear(hk.Module):
    """ Basic equivariant Linear layer from repin to repout."""
    def __init__(self, Pw,Pb,shape,name=None):
        super().__init__(name=name)
        self.Pw = Pw
        self.Pb = Pb
        self.shape=shape

    def __call__(self, x):  # (cin) -> (cout)
        i,j = self.shape
        w_init = hk.initializers.TruncatedNormal(1. / np.sqrt(i))
        w = hk.get_parameter("w", shape=self.shape, dtype=x.dtype, init=w_init)
        b = hk.get_parameter("b", shape=[i], dtype=x.dtype, init=w_init)
        W = (self.Pw@w.reshape(-1)).reshape(*self.shape)
        b = self.Pb@b
        return x@W.T+b

@export
def BiLinear(repin,repout):
    """ Cheap bilinear layer (adds parameters for each part of the input which can be
        interpreted as a linear map from a part of the input to the output representation)."""
    Wdim, weight_proj = bilinear_weights(repout,repin)
    return lambda x: hkBiLinear(weight_proj,Wdim)(x)


class hkBiLinear(hk.Module):
    def __init__(self, weight_proj,Wdim,name=None):
        super().__init__(name=name)
        self.weight_proj=weight_proj
        self.Wdim=Wdim

    def __call__(self, x):
        # compatible with non sumreps? need to check
        w_init = hk.initializers.TruncatedNormal(1.)
        w = hk.get_parameter("w", shape=[self.Wdim], dtype=x.dtype, init=w_init)
        W = self.weight_proj(w,x)
        return .1*(W@x[...,None])[...,0]

@export
class GatedNonlinearity(object):  # TODO: add support for mixed tensors and non sumreps
    """ Gated nonlinearity. Requires input to have the additional gate scalars
        for every non regular and non scalar rep. Applies swish to regular and
        scalar reps. (Right now assumes rep is a SumRep)"""
    def __init__(self,rep,name=None):
        super().__init__()
        self.rep=rep

    def __call__(self,values):
        gate_scalars = values[..., gate_indices(self.rep)]
        activations = jax.nn.sigmoid(gate_scalars) * values[..., :self.rep.size()]
        return activations


@export
def EMLPBlock(repin,repout):
    """ Basic building block of EMLP consisting of G-Linear, biLinear,
        and gated nonlinearity. """
    linear = Linear(repin,gated(repout))
    bilinear = BiLinear(gated(repout),gated(repout))
    nonlinearity = GatedNonlinearity(repout)
    def block(x):
        lin = linear(x)
        preact =bilinear(lin)+lin
        return nonlinearity(preact)
    return block


@export
def EMLP(rep_in,rep_out,group,ch=384,num_layers=3):
    """ Equivariant MultiLayer Perceptron.
        If the input ch argument is an int, uses the hands off uniform_rep heuristic.
        If the ch argument is a representation, uses this representation for the hidden layers.
        Individual layer representations can be set explicitly by using a list of ints or a list of
        representations, rather than use the same for each hidden layer.

        Args:
            rep_in (Rep): input representation
            rep_out (Rep): output representation
            group (Group): symmetry group
            ch (int or list[int] or Rep or list[Rep]): number of channels in the hidden layers
            num_layers (int): number of hidden layers

        Returns:
            Module: the EMLP objax module."""
    logging.info("Initing EMLP (Haiku)")
    rep_in =rep_in(group)
    rep_out = rep_out(group)
    # Parse ch as a single int, a sequence of ints, a single Rep, a sequence of Reps
    if isinstance(ch,int): middle_layers = num_layers*[uniform_rep(ch,group)]
    elif isinstance(ch,Rep): middle_layers = num_layers*[ch(group)]
    else: middle_layers = [(c(group) if isinstance(c,Rep) else uniform_rep(c,group)) for c in ch]
    # assert all((not rep.G is None) for rep in middle_layers[0].reps)
    reps = [rep_in]+middle_layers
    # logging.info(f"Reps: {reps}")
    network = Sequential(
        *[EMLPBlock(rin,rout) for rin,rout in zip(reps,reps[1:])],
        Linear(reps[-1],rep_out)
    )
    return network

@export
def MLP(rep_in,rep_out,group,ch=384,num_layers=3):
    """ Standard baseline MLP. Representations and group are used for shapes only. """
    cout = rep_out(group).size()
    mlp = lambda x: Sequential(
        *[Sequential(hk.Linear(ch),jax.nn.swish) for _ in range(num_layers)],
        hk.Linear(cout)
    )(x)
    return mlp



from .linear_operator_base import LinearOperator,Lazy
import jax.numpy as jnp
import numpy as np
from jax import jit
import jax
from functools import reduce

product = lambda c: reduce(lambda a,b:a*b,c)

def lazify(x):
    if isinstance(x,LinearOperator): return x
    elif isinstance(x,(jnp.ndarray,np.ndarray)): return Lazy(x)
    else: raise NotImplementedError

def densify(x):
    if isinstance(x,LinearOperator): return x.to_dense()
    elif isinstance(x,(jnp.ndarray,np.ndarray)): return x
    else: raise NotImplementedError

class I(LinearOperator):
    def __init__(self,d):
        shape = (d,d)
        super().__init__(None, shape)
    def _matmat(self,V): #(c,k)
        return V
    def _matvec(self,V):
        return V
    def _adjoint(self):
        return self
    def invT(self):
        return self

class LazyKron(LinearOperator):

    def __init__(self,Ms):
        self.Ms = Ms
        shape = product([Mi.shape[0] for Mi in Ms]), product([Mi.shape[1] for Mi in Ms])
        #self.dtype=Ms[0].dtype
        super().__init__(None,shape)

    def _matvec(self,v):
        return self._matmat(v).reshape(-1)
    def _matmat(self,v):
        ev = v.reshape(*[Mi.shape[-1] for Mi in self.Ms],-1)
        for i,M in enumerate(self.Ms):
            ev_front = jnp.moveaxis(ev,i,0)
            Mev_front = (M@ev_front.reshape(M.shape[-1],-1)).reshape(M.shape[0],*ev_front.shape[1:])
            ev = jnp.moveaxis(Mev_front,0,i)
        return ev.reshape(self.shape[0],ev.shape[-1])
    def _adjoint(self):
        return LazyKron([Mi.T for Mi in self.Ms])
    def invT(self):
        return LazyKron([M.invT() for M in self.Ms])
    def to_dense(self):
        Ms = [M.to_dense() if isinstance(M,LinearOperator) else M for M in self.Ms]
        return reduce(jnp.kron,Ms)
    def __new__(cls,Ms):
        if len(Ms)==1: return Ms[0]
        return super().__new__(cls)

#@jit
def kronsum(A,B):
    return jnp.kron(A,jnp.eye(B.shape[-1])) + jnp.kron(jnp.eye(A.shape[-1]),B)


class LazyKronsum(LinearOperator):
    
    def __init__(self,Ms):
        self.Ms = Ms
        shape = product([Mi.shape[0] for Mi in Ms]), product([Mi.shape[1] for Mi in Ms])
        #self.dtype=Ms[0].dtype
        dtype=jnp.dtype('float32')
        super().__init__(dtype,shape)

    def _matvec(self,v):
        return self._matmat(v).reshape(-1)

    def _matmat(self,v):
        ev = v.reshape(*[Mi.shape[-1] for Mi in self.Ms],-1)
        out = 0*ev
        for i,M in enumerate(self.Ms):
            ev_front = jnp.moveaxis(ev,i,0)
            Mev_front = (M@ev_front.reshape(M.shape[-1],-1)).reshape(M.shape[0],*ev_front.shape[1:])
            out += jnp.moveaxis(Mev_front,0,i)
        return out.reshape(self.shape[0],ev.shape[-1])
        
    def _adjoint(self):
        return LazyKronsum([Mi.T for Mi in self.Ms])
    def to_dense(self):
        Ms = [M.to_dense() if isinstance(M,LinearOperator) else M for M in self.Ms]
        return reduce(kronsum,Ms)
    def __new__(cls,Ms):
        if len(Ms)==1: return Ms[0]
        return super().__new__(cls)
## could also be implemented as follows, but the fusing the sum into a single linearOperator is faster
# def lazy_kronsum(Ms):
#     n = len(Ms)
#     lprod = np.cumprod([1]+[mi.shape[-1] for mi in Ms])
#     rprod = np.cumprod([1]+[mi.shape[-1] for mi in reversed(Ms)])[::-1]
#     return reduce(lambda a,b: a+b,[lazy_kron([I(lprod[i]),Mi,I(rprod[i+1])]) for i,Mi in enumerate(Ms)])


class LazyJVP(LinearOperator):
    def __init__(self,operator_fn,X,TX):
        self.shape = operator_fn(X).shape
        self.vjp = lambda v: jax.jvp(lambda x: operator_fn(x)@v,[X],[TX])[1]
        self.vjp_T = lambda v: jax.jvp(lambda x: operator_fn(x).T@v,[X],[TX])[1]
        self.dtype=jnp.dtype('float32')
    def _matmat(self,v):
        return self.vjp(v)
    def _matvec(self,v):
        return self.vjp(v)
    def _rmatmat(self,v):
        return self.vjp_T(v)
    

class ConcatLazy(LinearOperator):
    """ Produces a linear operator equivalent to concatenating
        a collection of matrices Ms along axis=0 """
    def __init__(self,Ms):
        self.Ms = Ms
        assert all(M.shape[0]==Ms[0].shape[0] for M in Ms),\
             f"Trying to concatenate matrices of different sizes {[M.shape for M in Ms]}"
        shape = (sum(M.shape[0] for M in Ms),Ms[0].shape[1])
        super().__init__(None,shape)

    def _matmat(self,V):
        return jnp.concatenate([M@V for M in self.Ms],axis=0)
    def _rmatmat(self,V):
        Vs = jnp.split(V,len(self.Ms))
        return sum([self.Ms[i].T@Vs[i] for i in range(len(self.Ms))])
    def to_dense(self):
        dense_Ms = [M.to_dense() if isinstance(M,LinearOperator) else M for M in self.Ms]
        return jnp.concatenate(dense_Ms,axis=0)
    
class LazyDirectSum(LinearOperator):
    def __init__(self,Ms,multiplicities=None):
        self.Ms = [jax.device_put(M.astype(np.float32)) if isinstance(M,(np.ndarray)) else M for M in Ms]
        self.multiplicities = [1 for M in Ms] if multiplicities is None else multiplicities
        shape = (sum(Mi.shape[0]*c for Mi,c in zip(Ms,multiplicities)),
                      sum(Mi.shape[0]*c for Mi,c in zip(Ms,multiplicities)))
        super().__init__(None,shape)
        #self.dtype=Ms[0].dtype
        #self.dtype=jnp.dtype('float32')

    def _matvec(self,v):
        return lazy_direct_matmat(v,self.Ms,self.multiplicities)

    def _matmat(self,v): # (n,k)
        return lazy_direct_matmat(v,self.Ms,self.multiplicities)
    def _adjoint(self):
        return LazyDirectSum([Mi.T for Mi in self.Ms])
    def invT(self):
        return LazyDirectSum([M.invT() for M in self.Ms])
    def to_dense(self):
        Ms_all = [M for M,c in zip(self.Ms,self.multiplicities) for _ in range(c)]
        Ms_all = [Mi.to_dense() if isinstance(Mi,LinearOperator) else Mi for Mi in Ms_all]
        return jax.scipy.linalg.block_diag(*Ms_all)
    # def __new__(cls,Ms,multiplicities=None):
    #     if len(Ms)==1 and multiplicities is None: return Ms[0]
    #     return super().__new__(cls)
        
def lazy_direct_matmat(v,Ms,mults):
    n = v.shape[0]
    k = v.shape[1] if len(v.shape)>1 else 1
    i=0
    y = []
    for M, multiplicity in zip(Ms,mults):
        i_end = i+multiplicity*M.shape[-1]
        elems = M@v[i:i_end].T.reshape(k*multiplicity,M.shape[-1]).T
        y.append(elems.T.reshape(k,multiplicity*M.shape[0]).T)
        i = i_end
    y = jnp.concatenate(y,axis=0) #concatenate over rep axis
    return  y


class LazyPerm(LinearOperator):
    def __init__(self,perm):
        self.perm=perm
        shape = (len(perm),len(perm))
        super().__init__(None,shape)

    def _matmat(self,V):
        return V[self.perm]
    def _matvec(self,V):
        return V[self.perm]
    def _adjoint(self):
        return LazyPerm(np.argsort(self.perm))
    def invT(self):
        return self

class LazyShift(LinearOperator):
    def __init__(self,n,k=1):
        self.k=k
        shape = (n,n)
        super().__init__(None,shape)

    def _matmat(self,V): #(c,k) #Still needs to be tested??
        return jnp.roll(V,self.k,axis=0)
    def _matvec(self,V):
        return jnp.roll(V,self.k,axis=0)
    def _adjoint(self):
        return LazyShift(self.shape[0],-self.k)
    def invT(self):
        return self

class SwapMatrix(LinearOperator):
    def __init__(self,swaprows,n):
        self.swaprows=swaprows
        shape = (n,n)
        super().__init__(None,shape)
    def _matmat(self,V): #(c,k)
        V = jax.ops.index_update(V, jax.ops.index[self.swaprows], V[self.swaprows[::-1]])
        return V
    def _matvec(self,V):
        return self._matmat(V)
    def _adjoint(self):
        return self
    def invT(self):
        return self

class Rot90(LinearOperator):
    def __init__(self,n,k):
        shape = (n*n,n*n)
        self.n=n
        self.k = k
        super().__init__(None,shape)
    def _matmat(self,V): #(c,k)
        return jnp.rot90(V.reshape((self.n,self.n,-1)),self.k).reshape(V.shape)
    def _matvec(self,V):
        return jnp.rot90(V.reshape((self.n,self.n,-1)),self.k).reshape(V.shape)
    def invT(self):
        return self


import numpy as np
import jax
import jax.numpy as jnp
from jax import jit, device_put,vmap
import optax
from sklearn.cluster import KMeans
from tqdm.auto import tqdm
from .linear_operator_base import LinearOperator, Lazy
from .linear_operators import ConcatLazy, I, lazify, densify, LazyJVP
import logging
import matplotlib.pyplot as plt
from functools import reduce
from emlp.utils import export

from plum import dispatch
import emlp.reps
#TODO: add rep,v = flatten({'Scalar':..., 'Vector':...,}), to_dict(rep,vector) returns {'Scalar':..., 'Vector':...,}
#TODO and simpler rep = flatten({Scalar:2,Vector:10,...}),
# Do we even want + operator to implement non canonical orderings?

__all__ = ["V","Vector", "Scalar"]

@export
class Rep(object):
    r""" The base Representation class. Representation objects formalize the vector space V
        on which the group acts, the group representation matrix ρ(g), and the Lie Algebra
        representation dρ(A) in a single object. Representations act as types for vectors coming
        from V. These types can be manipulated and transformed with the built in operators
        ⊕,⊗,dual, as well as incorporating custom representations. Rep objects should
        be immutable.

        At minimum, new representations need to implement ``rho``, ``__str__``."""
        
    is_permutation=False

    def rho(self,M):
        """ Group representation of the matrix M of shape (d,d)"""
        raise NotImplementedError
        
    def drho(self,A): 
        """ Lie Algebra representation of the matrix A of shape (d,d)"""
        In = jnp.eye(A.shape[0])
        return LazyJVP(self.rho,In,A)

    def __call__(self,G):
        """ Instantiate (non concrete) representation with a given symmetry group"""
        raise NotImplementedError

    def __str__(self): raise NotImplementedError 
    #TODO: separate __repr__ and __str__?
    def __repr__(self): return str(self)
    
    
    def __eq__(self,other):
        if type(self)!=type(other): return False
        d1 = tuple([(k,v) for k,v in self.__dict__.items() if (k not in ['_size','is_permutation','is_orthogonal'])])
        d2 = tuple([(k,v) for k,v in other.__dict__.items() if (k not in ['_size','is_permutation','is_orthogonal'])])
        return d1==d2
    def __hash__(self):
        d1 = tuple([(k,v) for k,v in self.__dict__.items() if (k not in ['_size','is_permutation','is_orthogonal'])])
        return hash((type(self),d1))

    def size(self): 
        """ Dimension dim(V) of the representation """
        if hasattr(self,'_size'):
            return self._size
        elif self.concrete and hasattr(self,"G"):
            self._size = self.rho(self.G.sample()).shape[-1]
            return self._size
        else: raise NotImplementedError

    def canonicalize(self): 
        """ An optional method to convert the representation into a canonical form
            in order to reuse equivalent solutions in the solver. Should return
            both the canonically ordered representation, along with a permutation
            which can be applied to vectors of the current representation to achieve
            that ordering. """
        return self, np.arange(self.size()) # return canonicalized rep
    
    def rho_dense(self,M):
        """ A convenience function which returns rho(M) as a dense matrix."""
        return densify(self.rho(M))
    def drho_dense(self,A):
        """ A convenience function which returns drho(A) as a dense matrix."""
        return densify(self.drho(A))
    
    def constraint_matrix(self):
        """ Constructs the equivariance constrant matrix (lazily) by concatenating
        the constraints (ρ(hᵢ)-I) for i=1,...M and dρ(Aₖ) for k=1,..,D from the generators
        of the symmetry group. """
        n = self.size()
        constraints = []
        constraints.extend([lazify(self.rho(h))-I(n) for h in self.G.discrete_generators])
        constraints.extend([lazify(self.drho(A)) for A in self.G.lie_algebra])
        return ConcatLazy(constraints) if constraints else lazify(jnp.zeros((1,n)))

    solcache = {}
    def equivariant_basis(self):  
        """ Computes the equivariant solution basis for the given representation of size N.
            Canonicalizes problems and caches solutions for reuse. Output [Q (N,r)] """
        if self==Scalar: return jnp.ones((1,1))
        canon_rep,perm = self.canonicalize()
        invperm = np.argsort(perm)
        if canon_rep not in self.solcache:
            logging.info(f"{canon_rep} cache miss")
            logging.info(f"Solving basis for {self}"+(f", for G={self.G}" if hasattr(self,"G") else ""))
            #if isinstance(group,Trivial): return np.eye(size(rank,group.d))
            C_lazy = canon_rep.constraint_matrix()
            if C_lazy.shape[0]*C_lazy.shape[1]>3e7: #Too large to use SVD
                result = krylov_constraint_solve(C_lazy)
            else:
                C_dense = C_lazy.to_dense()
                result = orthogonal_complement(C_dense)
            self.solcache[canon_rep]=result
        return self.solcache[canon_rep][invperm]
    
    def equivariant_projector(self):
        """ Computes the (lazy) projection matrix P=QQᵀ that projects to the equivariant basis."""
        Q = self.equivariant_basis()
        Q_lazy = lazify(Q)
        P = Q_lazy@Q_lazy.H
        return P

    @property
    def concrete(self):
        return hasattr(self,"G") and self.G is not None
        # if hasattr(self,"_concrete"): return self._concrete
        # else:
        #     return hasattr(self,"G") and self.G is not None

    def __add__(self, other):
        """ Direct sum (⊕) of representations. """
        if isinstance(other,int):
            if other==0: return self
            else: return self+other*Scalar
        elif emlp.reps.product_sum_reps.both_concrete(self,other):
            return emlp.reps.product_sum_reps.SumRep(self,other)
        else:
            return emlp.reps.product_sum_reps.DeferredSumRep(self,other)

    def __radd__(self,other):
        if isinstance(other,int): 
            if other==0: return self
            else: return other*Scalar+self
        else: return NotImplemented
        
    def __mul__(self,other):
        """ Tensor sum (⊗) of representations. """
        return mul_reps(self,other)
            
    def __rmul__(self,other):
        return mul_reps(other,self)

    def __pow__(self,other):
        """ Iterated tensor product. """
        assert isinstance(other,int), f"Power only supported for integers, not {type(other)}"
        assert other>=0, f"Negative powers {other} not supported"
        return reduce(lambda a,b:a*b,other*[self],Scalar)
    def __rshift__(self,other):
        """ Linear maps from self -> other """
        return other*self.T
    def __lshift__(self,other):
        """ Linear maps from other -> self """
        return self*other.T
    def __lt__(self, other):
        """ less than defined to disambiguate ordering multiple different representations.
            Canonical ordering is determined first by Group, then by size, then by hash"""
        if other==Scalar: return False
        try: 
            if self.G<other.G: return True
            if self.G>other.G: return False
        except (AttributeError,TypeError): pass
        if self.size()<other.size(): return True
        if self.size()>other.size(): return False
        return hash(self) < hash(other) #For sorting purposes only
    def __mod__(self,other): # Wreath product
        """ Wreath product of representations (Not yet implemented)"""
        raise NotImplementedError
    @property
    def T(self):
        """ Dual representation V*, rho*, drho*."""
        if hasattr(self,"G") and (self.G is not None) and self.G.is_orthogonal: return self
        return Dual(self)


@dispatch
def mul_reps(ra,rb:int):
    if rb==1: return ra
    if rb==0: return 0
    if (not hasattr(ra,'concrete')) or ra.concrete:
        return emlp.reps.product_sum_reps.SumRep(*(rb*[ra]))
    else:
        return emlp.reps.product_sum_reps.DeferredSumRep(*(rb*[ra]))

@dispatch
def mul_reps(ra:int,rb):
    return mul_reps(rb,ra)

# Continued with non int cases in product_sum_reps.py

# A possible
class ScalarRep(Rep):
    def __init__(self,G=None):
        self.G=G
        self.is_permutation = True
    def __call__(self,G):
        self.G=G
        return self
    def size(self):
        return 1
    def __repr__(self): return str(self)#f"T{self.rank+(self.G,)}"
    def __str__(self):
        return "V⁰"
    @property
    def T(self):
        return self
    def rho(self,M):
        return jnp.eye(1)
    def drho(self,M):
        return 0*jnp.eye(1)
    def __hash__(self):
        return 0
    def __eq__(self,other):
        return isinstance(other,ScalarRep)
    def __mul__(self,other):
        if isinstance(other,int): return super().__mul__(other)
        return other
    def __rmul__(self,other):
        if isinstance(other,int): return super().__rmul__(other)
        return other
    @property
    def concrete(self):
        return True

class Base(Rep):
    """ Base representation V of a group."""
    def __init__(self,G=None):
        self.G=G
        if G is not None: self.is_permutation = G.is_permutation
    def __call__(self,G):
        return self.__class__(G)
    def rho(self,M):
        if hasattr(self,'G') and isinstance(M,dict): M=M[self.G]
        return M
    def drho(self,A):
        if hasattr(self,'G') and isinstance(A,dict): A=A[self.G]
        return A
    def size(self):
        assert self.G is not None, f"must know G to find size for rep={self}"
        return self.G.d
    def __repr__(self): return str(self)#f"T{self.rank+(self.G,)}"
    def __str__(self):
        return "V"# +(f"_{self.G}" if self.G is not None else "")
    
    def __hash__(self):
        return hash((type(self),self.G))
    def __eq__(self,other):
        return type(other)==type(self) and self.G==other.G
    def __lt__(self,other):
        if isinstance(other,Dual): return True
        return super().__lt__(other)
    # @property
    # def T(self):
    #     return Dual(self.G)

class Dual(Rep):
    def __init__(self,rep):
        self.rep = rep
        self.G=rep.G
        if hasattr(rep,"is_permutation"): self.is_permutation = rep.is_permutation
    def __call__(self,G):
        return self.rep(G).T
    def rho(self,M):
        rho = self.rep.rho(M)
        rhoinvT = rho.invT() if isinstance(rho,LinearOperator) else jnp.linalg.inv(rho).T
        return rhoinvT
    def drho(self,A):
        return -self.rep.drho(A).T
    def __str__(self):
        return str(self.rep)+"*"
    def __repr__(self): return str(self)
    @property
    def T(self):
        return self.rep
    def __eq__(self,other):
        return type(other)==type(self) and self.rep==other.rep
    def __hash__(self):
        return hash((type(self),self.rep))
    def __lt__(self,other):
        if other==self.rep: return False
        return super().__lt__(other)
    def size(self):
        return self.rep.size()
        
V=Vector= Base()  #: Alias V or Vector for an instance of the Base representation of a group

Scalar = ScalarRep()#: An instance of the Scalar representation, equivalent to V**0

@export
def T(p,q=0,G=None):
    """ A convenience function for creating rank (p,q) tensors."""
    return (V**p*V.T**q)(G)

def orthogonal_complement(proj):
    """ Computes the orthogonal complement to a given matrix proj"""
    U,S,VH = jnp.linalg.svd(proj,full_matrices=True)
    rank = (S>1e-5).sum()
    return VH[rank:].conj().T

def krylov_constraint_solve(C,tol=1e-5):
    """ Computes the solution basis Q for the linear constraint CQ=0  and QᵀQ=I
        up to specified tolerance with C expressed as a LinearOperator. """
    r = 5
    if C.shape[0]*r*2>2e9: raise Exception(f"Solns for contraints {C.shape} too large to fit in memory")
    found_rank=5
    while found_rank==r:
        r *= 2 # Iterative doubling of rank until large enough to include the full solution space
        if C.shape[0]*r>2e9:
            logging.error(f"Hit memory limits, switching to sample equivariant subspace of size {found_rank}")
            break
        Q = krylov_constraint_solve_upto_r(C,r,tol)
        found_rank = Q.shape[-1]
    return Q

def krylov_constraint_solve_upto_r(C,r,tol=1e-5,lr=1e-2):#,W0=None):
    """ Iterative routine to compute the solution basis to the constraint CQ=0 and QᵀQ=I
        up to the rank r, with given tolerance. Uses gradient descent (+ momentum) on the
        objective |CQ|^2, which provably converges at an exponential rate."""
    W = np.random.randn(C.shape[-1],r)/np.sqrt(C.shape[-1])# if W0 is None else W0
    W = device_put(W)
    opt_init,opt_update = optax.sgd(lr,.9)
    opt_state = opt_init(W)  # init stats

    def loss(W):
        return (jnp.absolute(C@W)**2).sum()/2 # added absolute for complex support

    loss_and_grad = jit(jax.value_and_grad(loss))
    # setup progress bar
    pbar = tqdm(total=100,desc=f'Krylov Solving for Equivariant Subspace r<={r}',
    bar_format="{l_bar}{bar}| {n:.3g}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt}{postfix}]")
    prog_val = 0
    lstart, _ = loss_and_grad(W)
    
    for i in range(20000):
        
        lossval, grad = loss_and_grad(W)
        updates, opt_state = opt_update(grad, opt_state, W)
        W = optax.apply_updates(W, updates)
        # update progress bar
        progress = max(100*np.log(lossval/lstart)/np.log(tol**2/lstart)-prog_val,0)
        progress = min(100-prog_val,progress)
        if progress>0:
            prog_val += progress
            pbar.update(progress)

        if jnp.sqrt(lossval) <tol: # check convergence condition
            pbar.close()
            break # has converged
        if lossval>2e3 and i>100: # Solve diverged due to too high learning rate
            logging.warning(f"Constraint solving diverged, trying lower learning rate {lr/3:.2e}")
            if lr < 1e-4: raise ConvergenceError(f"Failed to converge even with smaller learning rate {lr:.2e}")
            return krylov_constraint_solve_upto_r(C,r,tol,lr=lr/3)
    else: raise ConvergenceError("Failed to converge.")
    # Orthogonalize solution at the end
    U,S,VT = np.linalg.svd(np.array(W),full_matrices=False) 
    # Would like to do economy SVD here (to not have the unecessary O(n^2) memory cost) 
    # but this is not supported in numpy (or Jax) unfortunately.
    rank = (S>10*tol).sum()
    Q = device_put(U[:,:rank])
    # final_L
    final_L = loss_and_grad(Q)[0]
    if final_L >tol: logging.warning(f"Normalized basis has too high error {final_L:.2e} for tol {tol:.2e}")
    scutoff = (S[rank] if r>rank else 0)
    assert rank==0 or scutoff < S[rank-1]/100, f"Singular value gap too small: {S[rank-1]:.2e} \
        above cutoff {scutoff:.2e} below cutoff. Final L {final_L:.2e}, earlier {S[rank-5:rank]}"
    #logging.debug(f"found Rank {r}, above cutoff {S[rank-1]:.3e} after {S[rank] if r>rank else np.inf:.3e}. Loss {final_L:.1e}")
    return Q

class ConvergenceError(Exception): pass

@export
def sparsify_basis(Q,lr=1e-2): #(n,r)
    """ Convenience function to attempt to sparsify a given basis by applying an orthogonal transformation
        W, Q' = QW where Q' has only 1s, 0s and -1s. Notably this method does not have the same convergence
        gauruntees of krylov_constraint_solve and can fail (even silently). Intended to be used only for
        visualization purposes, use at your own risk. """
    W = np.random.randn(Q.shape[-1],Q.shape[-1])
    W,_ = np.linalg.qr(W)
    W = device_put(W.astype(jnp.float32))
    opt_init,opt_update = optax.adam(lr)#optax.sgd(1e2,.9)#optax.adam(lr)#optax.sgd(3e-3,.9)#optax.adam(lr)
    opt_update = jit(opt_update)
    opt_state = opt_init(W)  # init stats

    def loss(W):
        return jnp.abs(Q@W.T).mean() + .1*(jnp.abs(W.T@W-jnp.eye(W.shape[0]))).mean()+.01*jax.numpy.linalg.slogdet(W)[1]**2

    loss_and_grad = jit(jax.value_and_grad(loss))

    for i in tqdm(range(3000),desc=f'sparsifying basis'):
        lossval, grad = loss_and_grad(W)
        updates, opt_state = opt_update(grad, opt_state, W)
        W = optax.apply_updates(W, updates)
        #W,_ = np.linalg.qr(W)
        if lossval>1e2 and i>100: # Solve diverged due to too high learning rate
            logging.warning(f"basis sparsification diverged, trying lower learning rate {lr/3:.2e}")
            return sparsify_basis(Q,lr=lr/3)
    Q = np.copy(Q@W.T)
    Q[np.abs(Q)<1e-2]=0
    Q[np.abs(Q)>1e-2] /= np.abs(Q[np.abs(Q)>1e-2])
    A = Q@(1+np.arange(Q.shape[-1]))
    if len(np.unique(np.abs(A)))!=Q.shape[-1]+1 and len(np.unique(np.abs(A)))!=Q.shape[-1]:
        logging.error(f"Basis elems did not separate: found only {len(np.unique(np.abs(A)))}/{Q.shape[-1]}")
        #raise ConvergenceError(f"Basis elems did not separate: found only {len(np.unique(A))}/{Q.shape[-1]}")
    return Q

#@partial(jit,static_argnums=(0,1))
@export
def bilinear_weights(out_rep,in_rep):
    #TODO: replace lazy_projection function with LazyDirectSum LinearOperator
    W_rep,W_perm = (in_rep>>out_rep).canonicalize()
    # possible bug when in_rep and out_rep are both non sumreps? #TODO: investigate
    inv_perm = np.argsort(W_perm)
    mat_shape = out_rep.size(),in_rep.size()
    x_rep=in_rep
    W_multiplicities = W_rep.reps
    x_multiplicities = x_rep.reps
    x_multiplicities = {rep:n for rep,n in x_multiplicities.items() if rep!=Scalar}
    nelems = lambda nx,rep: min(nx,rep.size())
    active_dims = sum([W_multiplicities.get(rep,0)*nelems(n,rep) for rep,n in x_multiplicities.items()])
    reduced_indices_dict = {rep:ids[np.random.choice(len(ids),nelems(len(ids),rep))].reshape(-1)\
                                for rep,ids in x_rep.as_dict(np.arange(x_rep.size())).items()}
    # Apply the projections for each rank, concatenate, and permute back to orig rank order
    @jit
    def lazy_projection(params,x): # (r,), (*c) #TODO: find out why backwards of this function is so slow
        bshape = x.shape[:-1]
        x = x.reshape(-1,x.shape[-1])
        bs = x.shape[0]
        i=0
        Ws = []
        for rep, W_mult in W_multiplicities.items():
            if rep not in x_multiplicities:
                Ws.append(jnp.zeros((bs,W_mult*rep.size())))
                continue
            x_mult = x_multiplicities[rep]
            n = nelems(x_mult,rep)
            i_end = i+W_mult*n
            bids =  reduced_indices_dict[rep]
            bilinear_params = params[i:i_end].reshape(W_mult,n) # bs,nK-> (nK,bs)
            i = i_end  # (bs,W_mult,d^r) = (W_mult,n)@(n,d^r,bs)
            bilinear_elems = bilinear_params@x[...,bids].T.reshape(n,rep.size()*bs)
            bilinear_elems = bilinear_elems.reshape(W_mult*rep.size(),bs).T
            Ws.append(bilinear_elems)
        Ws = jnp.concatenate(Ws,axis=-1) #concatenate over rep axis
        return Ws[...,inv_perm].reshape(*bshape,*mat_shape) # reorder to original rank ordering
    return active_dims,lazy_projection
        
# @jit
# def mul_part(bparams,x,bids):
#     b = prod(x.shape[:-1])
#     return (bparams@x[...,bids].T.reshape(bparams.shape[-1],-1)).reshape(-1,b).T

@export
def vis(repin,repout,cluster=True):
    """ A function to visualize the basis of equivariant maps repin>>repout
        as an image. Only use cluster=True if you know Pv will only have
        r distinct values (true for G<S(n) but not true for many continuous groups)."""
    rep = (repin>>repout)
    P = rep.equivariant_projector() # compute the equivariant basis
    Q = rep.equivariant_basis()
    v = np.random.randn(P.shape[1])  # sample random vector
    v = np.round(P@v,decimals=4)  # project onto equivariant subspace (and round)
    if cluster: # cluster nearby values for better color separation in plot
        v = KMeans(n_clusters=Q.shape[-1]).fit(v.reshape(-1,1)).labels_
    plt.imshow(v.reshape(repout.size(),repin.size()))
    plt.axis('off')


def scale_adjusted_rel_error(t1,t2,g):
    error = jnp.sqrt(jnp.mean(jnp.abs(t1-t2)**2))
    tscale = jnp.sqrt(jnp.mean(jnp.abs(t1)**2)) + jnp.sqrt(jnp.mean(jnp.abs(t2)**2))
    gscale = jnp.sqrt(jnp.mean(jnp.abs(g-jnp.eye(g.shape[-1]))**2))
    scale = jnp.maximum(tscale,gscale)
    return error/jnp.maximum(scale,1e-7)

@export
def equivariance_error(W,repin,repout,G):
    """ Computes the equivariance relative error rel_err(Wρ₁(g),ρ₂(g)W)
        of the matrix W (dim(repout),dim(repin)) [or basis Q: (dim(repout)xdim(repin), r)]
        according to the input and output representations and group G. """
    W = W.reshape(repout.size(),repin.size(),-1).transpose((2,0,1))[None]

    # Sample 5 group elements and verify the equivariance for each
    gs = G.samples(5)
    ring = vmap(repin.rho_dense)(gs)[:,None]
    routg = vmap(repout.rho_dense)(gs)[:,None]
    equiv_err = scale_adjusted_rel_error(W@ring,routg@W,gs)
    return equiv_err

import emlp.groups # Why is this necessary to avoid circular import?


import torch
from torch.autograd import Function
import jax
from jax import jit
from jax.tree_util import tree_flatten, tree_unflatten
import types
import copy
import jax.numpy as jnp
import numpy as np
import types
from functools import partial
from emlp.reps import T,Rep,Scalar
from emlp.reps import bilinear_weights
from emlp.utils import Named,export
import logging
import torch
import torch.nn as nn
import torch.nn.functional as F
from emlp.nn import gated,gate_indices,uniform_rep

def torch2jax(arr):
    if isinstance(arr,torch.Tensor):
        return jnp.asarray(arr.cpu().data.numpy())
    else:
        return arr

def jax2torch(arr):
    if isinstance(arr,(jnp.ndarray,np.ndarray)):
        if jax.devices()[0].platform=='gpu':
            device = torch.device('cuda') 
        else: device = torch.device('cpu')
        return torch.from_numpy(np.array(arr)).to(device)
    else:
        return arr

def to_jax(pytree):
    flat_values, tree_type = tree_flatten(pytree)
    transformed_flat = [torch2jax(v) for v in flat_values]
    return tree_unflatten(tree_type, transformed_flat)


def to_pytorch(pytree):
    flat_values, tree_type = tree_flatten(pytree)
    transformed_flat = [jax2torch(v) for v in flat_values]
    return tree_unflatten(tree_type, transformed_flat)

@export
def torchify_fn(function):
    """ A method to enable interopability between jax and pytorch autograd.
        Calling torchify on a given function that has pytrees of jax.ndarray
        objects as inputs and outputs will return a function that first converts
        the inputs to jax, runs through the jax function, and converts the output
        back to torch but preserving the gradients of the operation to be called
        with pytorch autograd. """
    vjp = jit(lambda *args: jax.vjp(function,*args))
    class torched_fn(Function):
        @staticmethod
        def forward(ctx,*args):
            if any(ctx.needs_input_grad):
                y,ctx.vjp_fn = vjp(*to_jax(args))#jax.vjp(function,*to_jax(args))
                return to_pytorch(y)
            return to_pytorch(function(*to_jax(args)))
        @staticmethod
        def backward(ctx,*grad_outputs):
            return to_pytorch(ctx.vjp_fn(*to_jax(grad_outputs)))
    return torched_fn.apply #TORCHED #Roasted


@export
class Linear(nn.Linear):
    """ Basic equivariant Linear layer from repin to repout."""
    def __init__(self, repin, repout):
        nin,nout = repin.size(),repout.size()
        super().__init__(nin,nout)
        rep_W = repout*repin.T
        rep_bias = repout
        Pw = rep_W.equivariant_projector()
        Pb = rep_bias.equivariant_projector()
        self.proj_b = torchify_fn(jit(lambda b: Pb@b))
        self.proj_w = torchify_fn(jit(lambda w:(Pw@w.reshape(-1)).reshape(nout,nin)))
        logging.info(f"Linear W components:{rep_W.size()} rep:{rep_W}")

    def forward(self, x): # (cin) -> (cout)
        return F.linear(x,self.proj_w(self.weight),self.proj_b(self.bias))

@export
class BiLinear(nn.Module):
    """ Cheap bilinear layer (adds parameters for each part of the input which can be
        interpreted as a linear map from a part of the input to the output representation)."""
    def __init__(self, repin, repout):
        super().__init__()
        Wdim, weight_proj = bilinear_weights(repout,repin)
        self.weight_proj = torchify_fn(jit(weight_proj))
        self.bi_params = nn.Parameter(torch.randn(Wdim))
        logging.info(f"BiW components: dim:{Wdim}")

    def forward(self, x,training=True):
        # compatible with non sumreps? need to check
        W = self.weight_proj(self.bi_params,x)
        out= .1*(W@x[...,None])[...,0]
        return out

@export
class GatedNonlinearity(nn.Module): #TODO: add support for mixed tensors and non sumreps
    """ Gated nonlinearity. Requires input to have the additional gate scalars
        for every non regular and non scalar rep. Applies swish to regular and
        scalar reps. (Right now assumes rep is a SumRep)"""
    def __init__(self,rep):
        super().__init__()
        self.rep=rep
    def forward(self,values):
        gate_scalars = values[..., gate_indices(self.rep)]
        activations = gate_scalars.sigmoid() * values[..., :self.rep.size()]
        return activations

@export
class EMLPBlock(nn.Module):
    """ Basic building block of EMLP consisting of G-Linear, biLinear,
        and gated nonlinearity. """
    def __init__(self,rep_in,rep_out):
        super().__init__()
        self.linear = Linear(rep_in,gated(rep_out))
        self.bilinear = BiLinear(gated(rep_out),gated(rep_out))
        self.nonlinearity = GatedNonlinearity(rep_out)

    def forward(self,x):
        lin = self.linear(x)
        preact =self.bilinear(lin)+lin
        return self.nonlinearity(preact)

@export
class EMLP(nn.Module):
    """ Equivariant MultiLayer Perceptron. 
        If the input ch argument is an int, uses the hands off uniform_rep heuristic.
        If the ch argument is a representation, uses this representation for the hidden layers.
        Individual layer representations can be set explicitly by using a list of ints or a list of
        representations, rather than use the same for each hidden layer.

        Args:
            rep_in (Rep): input representation
            rep_out (Rep): output representation
            group (Group): symmetry group
            ch (int or list[int] or Rep or list[Rep]): number of channels in the hidden layers
            num_layers (int): number of hidden layers

        Returns:
            Module: the EMLP objax module."""
    def __init__(self,rep_in,rep_out,group,ch=384,num_layers=3):
        super().__init__()
        logging.info("Initing EMLP (PyTorch)")
        self.rep_in =rep_in(group)
        self.rep_out = rep_out(group)
        
        self.G=group
        # Parse ch as a single int, a sequence of ints, a single Rep, a sequence of Reps
        if isinstance(ch,int): middle_layers = num_layers*[uniform_rep(ch,group)]#[uniform_rep(ch,group) for _ in range(num_layers)]
        elif isinstance(ch,Rep): middle_layers = num_layers*[ch(group)]
        else: middle_layers = [(c(group) if isinstance(c,Rep) else uniform_rep(c,group)) for c in ch]
        #assert all((not rep.G is None) for rep in middle_layers[0].reps)
        reps = [self.rep_in]+middle_layers
        #logging.info(f"Reps: {reps}")
        self.network = nn.Sequential(
            *[EMLPBlock(rin,rout) for rin,rout in zip(reps,reps[1:])],
            Linear(reps[-1],self.rep_out)
        )
    def forward(self,x):
        return self.network(x)

class Swish(nn.Module):
    def forward(self,x):
        return x.sigmoid()*x

def MLPBlock(cin,cout):
    return nn.Sequential(nn.Linear(cin,cout),Swish())#,nn.BatchNorm0D(cout,momentum=.9),swish)#,

@export
class MLP(nn.Module):
    """ Standard baseline MLP. Representations and group are used for shapes only. """
    def __init__(self,rep_in,rep_out,group,ch=384,num_layers=3):
        super().__init__()
        self.rep_in =rep_in(group)
        self.rep_out = rep_out(group)
        self.G = group
        chs = [self.rep_in.size()] + num_layers*[ch]
        cout = self.rep_out.size()
        logging.info("Initing MLP")
        self.net = nn.Sequential(
            *[MLPBlock(cin,cout) for cin,cout in zip(chs,chs[1:])],
            nn.Linear(chs[-1],cout)
        )

    def forward(self,x):
        y = self.net(x)
        return y

@export
class Standardize(nn.Module):
    """ A convenience module to wrap a given module, normalize its input
        by some dataset x mean and std stats, and unnormalize its output by
        the dataset y mean and std stats. 

        Args:
            model (Module): model to wrap
            ds_stats ((μx,σx,μy,σy) or (μx,σx)): tuple of the normalization stats
        
        Returns:
            Module: Wrapped model with input normalization (and output unnormalization)"""
    def __init__(self,model,ds_stats):
        super().__init__()
        self.model = model
        self.ds_stats=ds_stats

    def forward(self,x,training):
        if len(self.ds_stats)==2:
            muin,sin = self.ds_stats
            return self.model((x-muin)/sin,training=training)
        else:
            muin,sin,muout,sout = self.ds_stats
            y = sout*self.model((x-muin)/sin,training=training)+muout
            return y



from setuptools import setup,find_packages
import sys, os, re

README_FILE = 'README.md'

def get_property(prop, project):
    result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), open(project + '/__init__.py').read())
    return result.group(1)

project_name = "emlp"
setup(name=project_name,
      description="A Practical Method for Constructing Equivariant Multilayer Perceptrons for Arbitrary Matrix Groups",
      version= get_property('__version__',project_name),
      author='Marc Finzi',
      author_email='maf820@nyu.edu',
      license='MIT',
      python_requires='>=3.6',
      install_requires=['h5py','objax','pytest','plum-dispatch','optax','tqdm>=4.38'],
      extras_require = {
          'EXPTS':['olive-oil-ml']
      },
      packages=find_packages(),
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      url='https://github.com/mfinzi/equivariant-MLP',
      classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
      ],
      keywords=[
            'equivariance','MLP','symmetry','group','AI','neural network',
            'representation','group theory','deep learning','machine learning',
            'rotation','Lorentz invariance',
      ],

)



import numpy as np
import jax
from jax import jit
import collections,itertools
from functools import lru_cache as cache
from .representation import Rep,ScalarRep,Scalar
from .linear_operator_base import LinearOperator
from .linear_operators import LazyPerm,LazyDirectSum,LazyKron,LazyKronsum,I,lazy_direct_matmat,lazify,product
from functools import reduce
from collections import defaultdict
from plum import dispatch

class SumRep(Rep):
    def __init__(self,*reps,extra_perm=None):#repcounter,repperm=None):
        """ Constructs a tensor type based on a list of tensor ranks
            and possibly the symmetry generators gen."""
        # Integers can be used as shorthand for scalars.
        reps = [SumRepFromCollection({Scalar:rep}) if isinstance(rep,int) else rep for rep in reps]
        # Get reps and permutations
        reps,perms = zip(*[rep.canonicalize() for rep in reps])
        rep_counters = [rep.reps if isinstance(rep,SumRep) else {rep:1} for rep in reps]
        # Combine reps and permutations: ∑_a + ∑_b = ∑_{a∪b}
        #self.reps = sum(rep_counters,Counter({}))
        self.reps,perm = self.compute_canonical(rep_counters,perms)
        self.perm = extra_perm[perm] if extra_perm is not None else perm
        self.invperm = np.argsort(self.perm)
        self.canonical=(self.perm==np.arange(len(self.perm))).all()
        self.is_permutation = all(rep.is_permutation for rep in self.reps.keys())

    def size(self):
        return sum(rep.size()*count for rep,count in self.reps.items())

    def rho(self,M):
        rhos = [rep.rho(M) for rep in self.reps]
        multiplicities = self.reps.values()
        return LazyPerm(self.invperm)@LazyDirectSum(rhos,multiplicities)@LazyPerm(self.perm)

    def drho(self,A):
        drhos = [rep.drho(A) for rep in self.reps]
        multiplicities = self.reps.values()
        return LazyPerm(self.invperm)@LazyDirectSum(drhos,multiplicities)@LazyPerm(self.perm)

    def __eq__(self, other):
        return self.reps==other.reps and (self.perm==other.perm).all()

    def __hash__(self):
        assert self.canonical
        return hash(tuple(self.reps.items()))

    @property
    def T(self):
        """ only swaps to adjoint representation, does not reorder elems"""
        return SumRep(*[rep.T for rep,c in self.reps.items() for _ in range(c)],extra_perm=self.perm)
        # not necessarily still canonical ordered
        #return SumRepFromCollection({rep.T:c for rep,c in self.reps.items()},self.perm)

    def __repr__(self):
        return "+".join(f"{count if count > 1 else ''}{repr(rep)}" for rep, count in self.reps.items())
    def __str__(self):
        tensors = "+".join(f"{count if count > 1 else ''}{rep}" for rep, count in self.reps.items())
        return tensors#+f" @ d={self.d}" if self.d is not None else tensors

    def canonicalize(self):
        """Returns a canonically ordered rep with order np.arange(self.size()) and the
            permutation which achieves that ordering"""
        return SumRepFromCollection(self.reps),self.perm

    def __call__(self,G):
        return SumRepFromCollection({rep.T:c for rep,c in self.reps.items()},perm=self.perm)

    @property
    def concrete(self):
        return True

    def equivariant_basis(self):
        """ Overrides default implementation with a more efficient version which decomposes the constraints
            across the sum."""
        Qs = {rep: rep.equivariant_basis() for rep in self.reps}
        Qs = {rep: (jax.device_put(Q.astype(np.float32)) if isinstance(Q,(np.ndarray)) else Q) for rep,Q in Qs.items()}
        active_dims = sum([self.reps[rep]*Qs[rep].shape[-1] for rep in Qs.keys()])
        multiplicities = self.reps.values()
        def lazy_Q(array):
            return lazy_direct_matmat(array,Qs.values(),multiplicities)[self.invperm]
        return LinearOperator(shape=(self.size(),active_dims),matvec=lazy_Q,matmat=lazy_Q)

    def equivariant_projector(self):
        """ Overrides default implementation with a more efficient version which decomposes the constraints
            across the sum."""
        Ps = {rep:rep.equivariant_projector() for rep in self.reps}
        multiplicities = self.reps.values()
        def lazy_P(array):
            return lazy_direct_matmat(array[self.perm],Ps.values(),multiplicities)[self.invperm]#[:,self.invperm]
        return LinearOperator(shape=(self.size(),self.size()),matvec=lazy_P,matmat=lazy_P)

    # ##TODO: investigate why these more idiomatic definitions with Lazy Operators end up slower
    # def equivariant_basis(self):
    #     Qs = [rep.equivariant_basis() for rep in self.reps]
    #     Qs = [(jax.device_put(Q.astype(np.float32)) if isinstance(Q,(np.ndarray)) else Q) for Q in Qs]
    #     multiplicities  = self.reps.values()
    #     return LazyPerm(self.invperm)@LazyDirectSum(Qs,multiplicities)
    # def equivariant_projector(self):
    #     Ps = [rep.equivariant_projector() for rep in self.reps]
    #     Ps = (jax.device_put(P.astype(np.float32)) if isinstance(P,(np.ndarray)) else P)
    #     multiplicities  = self.reps.values()
    #     return LazyPerm(self.invperm)@LazyDirectSum(Ps,multiplicities)@LazyPerm(self.perm)

    # Some additional SumRep specific methods to be used for internal purposes
    @staticmethod
    def compute_canonical(rep_cnters,rep_perms):
        """ given that rep1_perm and rep2_perm are the canonical orderings for
            rep1 and rep2 (ie v[rep1_perm] is in canonical order) computes
            the canonical order for rep1 + rep2"""
        # First: merge counters
        unique_reps = sorted(reduce(lambda a,b: a|b,[cnter.keys() for cnter in rep_cnters]))
        merged_cnt = defaultdict(int)
        permlist = []
        ids = [0]*len(rep_cnters)
        shifted_perms = []
        n=0
        for perm in rep_perms:
            shifted_perms.append(n+perm)
            n+=len(perm)
        for rep in unique_reps:
            for i in range(len(ids)):
                c = rep_cnters[i].get(rep,0)
                permlist.append(shifted_perms[i][ids[i]:ids[i]+c*rep.size()])
                ids[i]+=+c*rep.size()
                merged_cnt[rep]+=c
        return dict(merged_cnt),np.concatenate(permlist)
    def __iter__(self):  # not a great idea to use this method (ignores permutation ordering)
        return (rep for rep,c in self.reps.items() for _ in range(c))
    def __len__(self):
        return sum(multiplicity for multiplicity in self.reps.values())
    def as_dict(self,v):
        out_dict = {}
        i =0
        for rep,c in self.reps.items():
            chunk = c*rep.size()
            out_dict[rep] = v[...,self.perm[i:i+chunk]].reshape(v.shape[:-1]+(c,rep.size()))
            i+= chunk
        return out_dict



def both_concrete(rep1,rep2):
    return all(rep.concrete for rep in (rep1,rep2) if hasattr(rep,'concrete'))

@dispatch.multi((SumRep,Rep),(Rep,SumRep),(SumRep,SumRep))
def mul_reps(ra,rb):
    if not both_concrete(ra,rb):
        return DeferredProductRep(ra,rb)
    return distribute_product([ra,rb])

@dispatch
def mul_reps(ra,rb):  # base case
    if type(ra) is ScalarRep: return rb
    if type(rb) is ScalarRep: return ra
    if not both_concrete(ra,rb):
        return DeferredProductRep(ra,rb)
    if hasattr(ra,"G") and hasattr(rb,"G") and ra.G==rb.G:
        return ProductRep(ra,rb)
    return DirectProduct(ra,rb)


    
#TODO: consolidate with the __init__ method of the basic SumRep
class SumRepFromCollection(SumRep): # a different constructor for SumRep
    def __init__(self,counter,perm=None):
        self.reps = counter
        self.perm = np.arange(self.size()) if perm is None else perm
        self.reps,self.perm = self.compute_canonical([counter],[self.perm])
        self.invperm = np.argsort(self.perm)
        self.canonical=(self.perm==np.arange(len(self.perm))).all()
        self.is_permutation = all(rep.is_permutation for rep in self.reps.keys())
        # if not self.canonical:
        #     print(self,self.perm,self.invperm)

def distribute_product(reps,extra_perm=None):
    """ For expanding products of sums into sums of products, (ρ₁⊕ρ₂)⊗ρ₃ = (ρ₁⊗ρ₃)⊕(ρ₂⊗ρ₃).
        takes in a sequence of reps=[ρ₁,ρ₂,ρ₃,...] which are to be multiplied together and at
        least one of the reps is a SumRep, and distributes out the terms."""
    reps,perms =zip(*[repsum.canonicalize() for repsum in reps])
    reps = [rep if isinstance(rep,SumRep) else SumRepFromCollection({rep:1}) for rep in reps]
    # compute axis_wise perm to canonical vector ordering along each axis
    
    axis_sizes = [len(perm) for perm in perms]

    order = np.arange(product(axis_sizes)).reshape(tuple(len(perm) for perm in perms))
    for i,perm in enumerate(perms):
        order = np.swapaxes(np.swapaxes(order,0,i)[perm,...],0,i)
    order = order.reshape(-1)
    #logging.info(f"axiswise: {order}")
    # Compute permutation from multilinear map ordering -> vector ordering (decomposing the blocks)
    repsizes_all = []
    for rep in reps:
        this_rep_sizes = []
        for r,c in rep.reps.items():
            this_rep_sizes.extend([c*r.size()])
        repsizes_all.append(tuple(this_rep_sizes))
    block_perm = rep_permutation(tuple(repsizes_all))
    #logging.info(f"block perm {block_perm}")
    # must go from itertools product ordering to multiplicity grouped ordering
    ordered_reps = []
    each_perm = []
    i = 0
    for prod in itertools.product(*[rep.reps.items() for rep in reps]):
        rs,cs = zip(*prod)
        #import pdb; pdb.set_trace()
        prod_rep,canonicalizing_perm = (product(cs)*reduce(lambda a,b: a*b,rs)).canonicalize()
        #print(f"{rs}:{cs} in distribute yield prod_rep {prod_rep}")
        ordered_reps.append(prod_rep)
        shape = []
        for r,c in prod:
            shape.extend([c,r.size()])
        axis_perm = np.concatenate([2*np.arange(len(prod)),2*np.arange(len(prod))+1])
        mul_perm = np.arange(len(canonicalizing_perm)).reshape(shape).transpose(axis_perm).reshape(-1)
        each_perm.append(mul_perm[canonicalizing_perm]+i)
        i+= len(canonicalizing_perm)
    each_perm = np.concatenate(each_perm)
    #logging.info(f"each perm {each_perm}")
    #
    total_perm = order[block_perm[each_perm]]
    if extra_perm is not None: total_perm = extra_perm[total_perm]
    #TODO: could achieve additional reduction by canonicalizing at this step, but unnecessary for now
    return SumRep(*ordered_reps,extra_perm=total_perm)


@cache(maxsize=None)
def rep_permutation(repsizes_all):
    """Permutation from block ordering to flattened ordering"""
    size_cumsums = [np.cumsum([0] + [size for size in repsizes]) for repsizes in repsizes_all]
    permutation = np.zeros([cumsum[-1] for cumsum in size_cumsums]).astype(int)
    arange = np.arange(permutation.size)
    indices_iter = itertools.product(*[range(len(repsizes)) for repsizes in repsizes_all])
    i = 0
    for indices in indices_iter:
        slices = tuple([slice(cumsum[idx], cumsum[idx + 1]) for idx, cumsum in zip(indices, size_cumsums)])
        slice_lengths = [sl.stop - sl.start for sl in slices]
        chunk_size = np.prod(slice_lengths)
        permutation[slices] += arange[i:i + chunk_size].reshape(*slice_lengths)
        i += chunk_size
    return np.argsort(permutation.reshape(-1))

class ProductRep(Rep):
    def __init__(self,*reps,extra_perm=None,counter=None):
        #Two variants of the constructor:
        if counter is not None: #one with counter specified directly
            self.reps=counter
            self.reps,self.perm = self.compute_canonical([counter],[np.arange(self.size()) if extra_perm is None else extra_perm])
        else: # other with list
            # Get reps and permutations
            reps,perms = zip(*[rep.canonicalize() for rep in reps])
            rep_counters = [rep.reps if type(rep)==ProductRep else {rep:1} for rep in reps]
            
            # Combine reps and permutations: ∏_a ⊗ ∏_b = ∏_{a ∪ b}
            self.reps,perm = self.compute_canonical(rep_counters,perms)
            self.perm = extra_perm[perm] if extra_perm is not None else perm

        self.invperm = np.argsort(self.perm)
        self.canonical=(self.perm==self.invperm).all()
        Gs = tuple(set(rep.G for rep in self.reps.keys()))
        assert len(Gs)==1, f"Multiple different groups {Gs} in product rep {self}"
        self.G= Gs[0]
        self.is_permutation = all(rep.is_permutation for rep in self.reps.keys())

    def size(self):
        return product([rep.size()**count for rep,count in self.reps.items()])

    def rho(self,Ms,lazy=False):
        if hasattr(self,'G') and isinstance(Ms,dict): Ms=Ms[self.G]
        canonical_lazy = LazyKron([rep.rho(Ms) for rep,c in self.reps.items() for _ in range(c)])
        return LazyPerm(self.invperm)@canonical_lazy@LazyPerm(self.perm)

    def drho(self,As):
        if hasattr(self,'G') and isinstance(As,dict): As=As[self.G]
        canonical_lazy = LazyKronsum([rep.drho(As) for rep,c in self.reps.items() for _ in range(c)])
        return LazyPerm(self.invperm)@canonical_lazy@LazyPerm(self.perm)

    def __hash__(self):
        assert self.canonical, f"Not canonical {repr(self)}? perm {self.perm}"
        return hash(tuple(self.reps.items()))
    def __eq__(self, other): #TODO: worry about non canonical?
        return isinstance(other,ProductRep) and self.reps==other.reps and (self.perm==other.perm).all()
    
    @property
    def concrete(self):
        return True

    @property
    def T(self):
        """ only swaps to adjoint representation, does not reorder elems"""
        return self.__class__(*[rep.T for rep,c in self.reps.items() for _ in range(c)],extra_perm=self.perm)
        #return self.__class__(counter={rep.T:c for rep,c in self.reps.items()},extra_perm=self.perm)
    def __str__(self):
        superscript = str.maketrans("0123456789", "⁰¹²³⁴⁵⁶⁷⁸⁹")
        return "⊗".join([str(rep)+(f"{c}".translate(superscript) if c>1 else "") for rep,c in self.reps.items()])

    def canonicalize(self):
        """Returns a canonically ordered rep with order np.arange(self.size()) and the
            permutation which achieves that ordering"""
        return self.__class__(counter=self.reps),self.perm

    @staticmethod
    def compute_canonical(rep_cnters,rep_perms):
        """ given that rep1_perm and rep2_perm are the canonical orderings for
            rep1 and rep2 (ie v[rep1_perm] is in canonical order) computes
            the canonical order for rep1 * rep2"""
        order = np.arange(product(len(perm) for perm in rep_perms))
        # First: merge counters
        unique_reps = sorted(reduce(lambda a,b: a|b,[cnter.keys() for cnter in rep_cnters]))
        merged_cnt = defaultdict(int)
        # Reshape like the tensor it is
        order = order.reshape(tuple(len(perm) for perm in rep_perms))
        # apply the canonicalizing permutations along each axis
        for i,perm in enumerate(rep_perms):
            order = np.moveaxis(np.moveaxis(order,i,0)[perm,...],0,i)
        # sort the axes by canonical ordering
        # get original axis ids
        axis_ids = []
        n=0
        for cnter in rep_cnters:
            axis_idsi = {}
            for rep,c in cnter.items():
                axis_idsi[rep] = n+np.arange(c)
                n+= c
            axis_ids.append(axis_idsi)
        axes_perm = []
        for rep in unique_reps:
            for i in range(len(rep_perms)):
                c = rep_cnters[i].get(rep,0)
                if c!=0:
                    axes_perm.append(axis_ids[i][rep])
                    merged_cnt[rep]+=c
        axes_perm = np.concatenate(axes_perm)
        # reshaped but with inner axes within a collection explicitly expanded
        order = order.reshape(tuple(rep.size() for cnter in rep_cnters for rep,c in cnter.items() for _ in range(c)))
        final_order = np.transpose(order,axes_perm)
        return dict(merged_cnt),final_order.reshape(-1)


class DirectProduct(ProductRep):
    """ Tensor product of representations ρ₁⊗ρ₂, but where the sub representations
        ρ₁ and ρ₂ are representations of distinct groups (ie ρ₁⊗ρ₂ is a representation
        of the direct product of groups G=G₁×G₂). As a result, the solutions for the two
        sub representations can be solved independently and assembled together with the
        kronecker product: Q = Q₁⊗Q₂ and P = P₁⊗P₂"""
    def __init__(self, *reps,counter=None,extra_perm=None):
        #Two variants of the constructor:
        if counter is not None: #one with counter specified directly
            self.reps=counter
            self.reps,perm = self.compute_canonical([counter],[np.arange(self.size())])
            self.perm = extra_perm[perm] if extra_perm is not None else perm
        else: # other with list
            reps,perms = zip(*[rep.canonicalize() for rep in reps])
            #print([type(rep) for rep in reps],type(rep1),type(rep2))
            rep_counters = [rep.reps if type(rep)==DirectProduct else {rep:1} for rep in reps]
            # Combine reps and permutations: Pi_a + Pi_b = Pi_{a x b}
            reps,perm = self.compute_canonical(rep_counters,perms)
            #print("dprod init",self.reps)
            group_dict = defaultdict(lambda: 1)
            for rep,c in reps.items():
                group_dict[rep.G]=group_dict[rep.G]*rep**c
            sub_products = {rep:1 for G,rep in group_dict.items()}
            self.reps = counter = sub_products
            self.reps,perm2 = self.compute_canonical([counter],[np.arange(self.size())])
            self.perm = extra_perm[perm[perm2]] if extra_perm is not None else perm[perm2]
        self.invperm = np.argsort(self.perm)
        self.canonical=(self.perm==self.invperm).all()
        # self.G = tuple(set(rep.G for rep in self.reps.keys()))
        # if len(self.G)==1: self.G= self.G[0]
        self.is_permutation = all(rep.is_permutation for rep in self.reps.keys())
        assert all(count==1 for count in self.reps.values())

    def equivariant_basis(self):
        canon_Q = LazyKron([rep.equivariant_basis() for rep,c in self.reps.items()])
        return LazyPerm(self.invperm)@canon_Q

    def equivariant_projector(self):
        canon_P = LazyKron([rep.equivariant_projector() for rep,c in self.reps.items()])
        return LazyPerm(self.invperm)@canon_P@LazyPerm(self.perm)

    def rho(self,Ms):
        canonical_lazy = LazyKron([rep.rho(Ms) for rep,c in self.reps.items() for _ in range(c)])
        return LazyPerm(self.invperm)@canonical_lazy@LazyPerm(self.perm)

    def drho(self,As):
        canonical_lazy = LazyKronsum([rep.drho(As) for rep,c in self.reps.items() for _ in range(c)])
        return LazyPerm(self.invperm)@canonical_lazy@LazyPerm(self.perm)

    def __str__(self):
        superscript = str.maketrans("0123456789", "⁰¹²³⁴⁵⁶⁷⁸⁹")
        return "⊗".join([str(rep)+f"_{rep.G}" for rep,c in self.reps.items()])


class DeferredSumRep(Rep):
    def __init__(self,*reps):
        self.to_sum=[]
        for rep in reps:
            #assert not isinstance(rep,SumRep),f"{rep} of type {type(rep)} tosum {self.to_sum}"
            self.to_sum.extend(rep.to_sum if isinstance(rep,DeferredSumRep) else [rep])
       
    def __call__(self,G):
        if G is None: return self
        return SumRep(*[rep(G) for rep in self.to_sum])
    def __repr__(self):
        return '('+"+".join(f"{rep}" for rep in self.to_sum)+')'
    def __str__(self):
        return repr(self)
    @property
    def T(self):
        return DeferredSumRep(*[rep.T for rep in self.to_sum])
    @property
    def concrete(self):
        return False

class DeferredProductRep(Rep):
    def __init__(self,*reps):
        self.to_prod=[]
        for rep in reps:
            assert not isinstance(rep,ProductRep)
            self.to_prod.extend(rep.to_prod if isinstance(rep,DeferredProductRep) else [rep])
    def __call__(self,G):
        if G is None: return self
        return reduce(lambda a,b:a*b,[rep(G) for rep in self.to_prod])
    def __repr__(self):
        return "⊗".join(f"{rep}" for rep in self.to_prod)
    def __str__(self):
        return repr(self)
    @property
    def T(self):
        return DeferredProductRep(*[rep.T for rep in self.to_prod])
    @property
    def concrete(self):
        return False



import numpy as np
from scipy.linalg import expm
from emlp.utils import Named,export
import jax
import jax.numpy as jnp
from emlp.reps.linear_operators import LazyShift,SwapMatrix,Rot90,LazyKron,LazyKronsum,LazyPerm,I
from jax import jit,vmap

def rel_err(A,B):
    return jnp.mean(jnp.abs(A-B))/(jnp.mean(jnp.abs(A)) + jnp.mean(jnp.abs(B))+1e-6)

@export
class Group(object,metaclass=Named):
    """ Abstract Group Object which new groups should inherit from. """
    lie_algebra = NotImplemented  #: The continuous generators
    discrete_generators = NotImplemented  #: The discrete generators
    z_scale=None # For scale noise for sampling elements
    is_orthogonal=None
    is_permutation = None
    d = NotImplemented  #: The dimension of the base representation
    def __init__(self,*args,**kwargs):
        # get the dimension of the base group representation
        if self.d is NotImplemented: 
            if self.lie_algebra is not NotImplemented and len(self.lie_algebra):
                self.d= self.lie_algebra[0].shape[-1]
            if self.discrete_generators is not NotImplemented and len(self.discrete_generators):
                self.d= self.discrete_generators[0].shape[-1]

        if self.lie_algebra is NotImplemented:
            self.lie_algebra = np.zeros((0,self.d,self.d))
        if self.discrete_generators is NotImplemented:
            self.discrete_generators = np.zeros((0,self.d,self.d))

        self.args = args
        if isinstance(self.lie_algebra,np.ndarray): self.lie_algebra = jax.device_put(self.lie_algebra)
        if isinstance(self.discrete_generators,np.ndarray): self.discrete_generators = jax.device_put(self.discrete_generators)

        # Set orthogonal flag automatically if not specified
        if self.is_permutation: self.is_orthogonal=True
        if self.is_orthogonal is None:
            self.is_orthogonal = True
            if len(self.lie_algebra)!=0:
                A_dense =jnp.stack([Ai@jnp.eye(self.d) for Ai in self.lie_algebra])
                self.is_orthogonal &= rel_err(-A_dense.transpose((0,2,1)),A_dense)<1e-6
            if len(self.discrete_generators)!=0:
                h_dense = jnp.stack([hi@jnp.eye(self.d) for hi in self.discrete_generators])
                self.is_orthogonal &= rel_err(h_dense.transpose((0,2,1))@h_dense,jnp.eye(self.d))<1e-6

        # Set regular flag automatically if not specified
        if self.is_orthogonal and (self.is_permutation is None):
            self.is_permutation=True
            self.is_permutation &= (len(self.lie_algebra)==0)  # no infinitesmal generators and all rows have one 1
            if len(self.discrete_generators)!=0:
                h_dense = jnp.stack([hi@jnp.eye(self.d) for hi in self.discrete_generators])
                self.is_permutation &= ((h_dense==1).astype(int).sum(-1)==1).all()

    def exp(self,A):
        """ Matrix exponential """
        return expm(A)

    def num_constraints(self):
        return len(self.lie_algebra)+len(self.discrete_generators)

    def sample(self):
        """Draw a sample from the group (not necessarily Haar measure)"""
        return self.samples(1)[0]

    def samples(self,N):
        """ Draw N samples from the group (not necessarily Haar measure)"""
        A_dense = jnp.stack([Ai@jnp.eye(self.d) for Ai in self.lie_algebra]) if len(self.lie_algebra) else jnp.zeros((0,self.d,self.d))
        h_dense = jnp.stack([hi@jnp.eye(self.d) for hi in self.discrete_generators]) if len(self.discrete_generators) else jnp.zeros((0,self.d,self.d))
        z = np.random.randn(N,A_dense.shape[0])
        if self.z_scale is not None:
            z*= self.z_scale
        k = np.random.randint(-5,5,size=(N,h_dense.shape[0],3))
        jax_seed=  np.random.randint(100)
        return noise2samples(z,k,A_dense,h_dense,jax_seed)

    def check_valid_group_elems(self,g):
        return True

    def __str__(self):
        return repr(self)

    def __repr__(self):
        outstr = f"{self.__class__}"
        if self.args:
            outstr += '('+''.join(repr(arg) for arg in self.args)+')'
        return outstr

    def __eq__(self,G2):  # TODO: more permissive by checking that spans are equal?
        return repr(self)==repr(G2)
        
    def __hash__(self):
        return hash(repr(self))

    def __lt__(self, other):
        """ For sorting purposes only """
        return hash(self) < hash(other)

    def __mul__(self,other):
        return DirectProduct(self,other)

@jit
def matrix_power_simple(M,n):
    out = jnp.eye(M.shape[-1])
    body = lambda Mn: jax.lax.fori_loop(0,Mn[1],lambda i,g: Mn[0]@g,out)
    out = jax.lax.cond(n<0,(jnp.linalg.inv(M),-n),body,(M,n),body)
    return out
    
@jit
def noise2sample(z,ks,lie_algebra,discrete_generators,seed=0):
    """ [zs (D,)] [ks (M,K)] [lie_algebra (D,d,d)] [discrete_generators (M,d,d)] 
        Here K is the number of repeats for a given discrete generator."""
    g = jnp.eye(lie_algebra.shape[-1])
    if lie_algebra.shape[0]:
        A = (z[:,None,None]*lie_algebra).sum(0)
        g = g@jax.scipy.linalg.expm(A)
    key = jax.random.PRNGKey(seed)
    M,K = ks.shape
    if M==0: return g
    for k in range(K): # multiple rounds of discrete generators
        key,pkey = jax.random.split(key)
        for i in jax.random.permutation(pkey,M): # Randomize the order of generators
            g = g@matrix_power_simple(discrete_generators[i],ks[i,k])#jnp.linalg.matrix_power(discrete_generators[i],ks[i])
    return g

@jit
def noise2samples(zs,ks,lie_algebra,discrete_generators,seed=0):
    return vmap(noise2sample,(0,0,None,None,None),0)(zs,ks,lie_algebra,discrete_generators,seed)

@export
class Trivial(Group):
    """ The trivial group G={I} in n dimensions. If you want to see how the
        inductive biases of EMLP perform without any symmetry, use Trivial(n)"""
    def __init__(self,n):
        self.d = n
        super().__init__(n)

@export
class SO(Group):
    """ The special orthogonal group SO(n) in n dimensions"""
    def __init__(self,n):
        self.lie_algebra = np.zeros(((n*(n-1))//2,n,n))
        k=0
        for i in range(n):
            for j in range(i):
                self.lie_algebra[k,i,j] = 1
                self.lie_algebra[k,j,i] = -1
                k+=1
        super().__init__(n)

@export
class O(SO):
    """ The Orthogonal group O(n) in n dimensions"""
    def __init__(self,n):
        self.discrete_generators = np.eye(n)[None]
        self.discrete_generators[0,0,0]=-1
        super().__init__(n)

@export
class C(Group):
    """ The Cyclic group Ck in 2 dimensions"""
    def __init__(self,k):
        theta = 2*np.pi/k
        self.discrete_generators = np.zeros((1,2,2))
        self.discrete_generators[0,:,:] = np.array([[np.cos(theta),np.sin(theta)],[-np.sin(theta),np.cos(theta)]])
        super().__init__(k)
@export
class D(C):
    """ The Dihedral group Dk in 2 dimensions"""
    def __init__(self,k):
        super().__init__(k)
        self.discrete_generators = np.concatenate((self.discrete_generators,np.array([[[-1,0],[0,1]]])))
@export
class Scaling(Group):
    """ The scaling group in n dimensions"""
    def __init__(self,n):
        self.lie_algebra = np.eye(n)[None]
        super().__init__(n)

class Parity(Group):  # """ The spacial parity group in 1+3 dimensions"""
    discrete_generators = -np.eye(4)[None]
    discrete_generators[0,0,0] = 1

class TimeReversal(Group):  # """ The time reversal group in 1+3 dimensions"""
    discrete_generators = np.eye(4)[None]
    discrete_generators[0,0,0] = -1

@export
class SO13p(Group):
    """ The component of Lorentz group connected to identity"""
    lie_algebra = np.zeros((6,4,4))
    lie_algebra[3:,1:,1:] = SO(3).lie_algebra
    for i in range(3):
        lie_algebra[i,1+i,0] = lie_algebra[i,0,1+i] = 1.

    # Adjust variance for samples along boost generators. For equivariance checks
    # the exps for high order tensors can get very large numbers
    z_scale = np.array([.3,.3,.3,1,1,1]) # can get rid of now

@export
class SO13(SO13p):
    discrete_generators = -np.eye(4)[None]

@export
class O13(SO13p):
    """ The full lorentz group (including Parity and Time reversal)"""
    discrete_generators = np.eye(4)[None] +np.zeros((2,1,1))
    discrete_generators[0] *= -1
    discrete_generators[1,0,0] = -1
@export
class Lorentz(O13): pass

@export
class SO11p(Group):
    """ The identity component of O(1,1) (Lorentz group in 1+1 dimensions)"""
    lie_algebra = np.array([[0.,1.],[1.,0.]])[None]

@export
class O11(SO11p):
    """ The Lorentz group O(1,1) in 1+1 dimensions """
    discrete_generators = np.eye(2)[None]+np.zeros((2,1,1))
    discrete_generators[0]*=-1
    discrete_generators[1,0,0] = -1

@export
class Sp(Group):
    """ Symplectic group Sp(m) in 2m dimensions (sometimes referred to
        instead as Sp(2m) )"""
    def __init__(self,m):
        self.lie_algebra = np.zeros((m*(2*m+1),2*m,2*m))
        k=0
        for i in range(m): # block diagonal elements
            for j in range(m):
                self.lie_algebra[k,i,j] = 1
                self.lie_algebra[k,m+j,m+i] = -1
                k+=1
        for i in range(m):
            for j in range(i+1):
                self.lie_algebra[k,m+i,j] = 1
                self.lie_algebra[k,m+j,i] = 1
                k+=1
                self.lie_algebra[k,i,m+j] = 1
                self.lie_algebra[k,j,m+i] = 1
                k+=1
        super().__init__(m)
   
@export
class Z(Group):
    r""" The cyclic group Z_n (discrete translation group) of order n.
        Features a regular base representation."""
    def __init__(self,n):
        self.discrete_generators = [LazyShift(n)]
        super().__init__(n)

@export
class S(Group): #The permutation group
    r""" The permutation group S_n with an n dimensional regular representation."""
    def __init__(self,n):
        # Here we choose n-1 generators consisting of swaps between the first element
        # and every other element
        perms = np.arange(n)[None]+np.zeros((n-1,1)).astype(int)
        perms[:,0] = np.arange(1,n)
        perms[np.arange(n-1),np.arange(1,n)[None]]=0
        self.discrete_generators = [LazyPerm(perm) for perm in perms]
        super().__init__(n)
        # We can also have chosen the 2 generator soln described in the paper, but
        # adding superflous extra generators surprisingly can sometimes actually *decrease* 
        # the runtime of the iterative krylov solver by improving the conditioning 
        # of the constraint matrix

@export
class SL(Group):
    """ The special linear group SL(n) in n dimensions"""
    def __init__(self,n):
        self.lie_algebra = np.zeros((n*n-1,n,n))
        k=0
        for i in range(n):
            for j in range(n):
                if i==j: continue #handle diag elements separately
                self.lie_algebra[k,i,j] = 1
                k+=1
        for l in range(n-1):
            self.lie_algebra[k,l,l] = 1
            self.lie_algebra[k,-1,-1] = -1
            k+=1
        super().__init__(n)

@export
class GL(Group):
    """ The general linear group GL(n) in n dimensions"""
    def __init__(self,n):
        self.lie_algebra = np.zeros((n*n,n,n))
        k=0
        for i in range(n):
            for j in range(n):
                self.lie_algebra[k,i,j] = 1
                k+=1
        super().__init__(n)

@export
class U(Group):  # Of dimension n^2
    """ The unitary group U(n) in n dimensions (complex)"""
    def __init__(self,n):
        lie_algebra_real = np.zeros((n**2,n,n))
        lie_algebra_imag = np.zeros((n**2,n,n))
        k=0
        for i in range(n):
            for j in range(i):
                # Antisymmetric real generators
                lie_algebra_real[k,i,j] = 1
                lie_algebra_real[k,j,i] = -1
                k+=1
                # symmetric imaginary generators
                lie_algebra_imag[k,i,j] = 1
                lie_algebra_imag[k,j,i] = 1
                k+=1
        for i in range(n):
            # diagonal imaginary generators
            lie_algebra_imag[k,i,i] = 1
            k+=1
        self.lie_algebra = lie_algebra_real + lie_algebra_imag*1j
        super().__init__(n)
@export
class SU(Group):  # Of dimension n^2-1
    """ The special unitary group SU(n) in n dimensions (complex)"""
    def __init__(self,n):
        if n==1: return Trivial(1)
        lie_algebra_real = np.zeros((n**2-1,n,n))
        lie_algebra_imag = np.zeros((n**2-1,n,n))
        k=0
        for i in range(n):
            for j in range(i):
                # Antisymmetric real generators
                lie_algebra_real[k,i,j] = 1
                lie_algebra_real[k,j,i] = -1
                k+=1
                # symmetric imaginary generators
                lie_algebra_imag[k,i,j] = 1
                lie_algebra_imag[k,j,i] = 1
                k+=1
        for i in range(n-1):
            # diagonal traceless imaginary generators
            lie_algebra_imag[k,i,i] = 1
            for j in range(n):
                if i==j: continue
                lie_algebra_imag[k,j,j] = -1/(n-1)
            k+=1
        self.lie_algebra = lie_algebra_real + lie_algebra_imag*1j
        super().__init__(n)

@export
class Cube(Group):
    """ A discrete version of SO(3) including all 90 degree rotations in 3d space
    Implements a 6 dimensional representation on the faces of a cube"""
    def __init__(self):
        #order = np.arange(6) # []
        Fperm = np.array([4,1,0,3,5,2])
        Lperm = np.array([3,0,2,5,4,1])
        self.discrete_generators = [LazyPerm(perm) for perm in [Fperm,Lperm]]
        super().__init__()


def pad(permutation):
    assert len(permutation)==48
    padded = np.zeros((6,9)).astype(permutation.dtype)
    padded[:,:4] = permutation.reshape(6,8)[:,:4]
    padded[:,5:] = permutation.reshape(6,8)[:,4:]
    return padded

def unpad(padded_perm):
    return np.concatenate([padded_perm[:,:4],padded_perm[:,5:]],-1).reshape(-1)




@export
class RubiksCube(Group): #3x3 rubiks cube
    r""" The Rubiks cube group G<S_48 consisting of all valid 3x3 Rubik's cube transformations.
        Generated by the a quarter turn about each of the faces."""
    def __init__(self):
        # Faces are ordered U,F,R,B,L,D (the net of the cube) #    B
        order = np.arange(48)                                 #  L U R
        order_padded = pad(order) # include a center element  #    F
        # Compute permutation for Up quarter turn             #    D
        order_padded[0,:] = np.rot90(order_padded[0].reshape(3,3),1).reshape(9) # Rotate top face
        FRBL = np.array([1,2,3,4])
        order_padded[FRBL,:3] = order_padded[np.roll(FRBL,1),:3] # F <- L,R <- F,B <- R,L <- B
        Uperm = unpad(order_padded)
        # Now form all other generators by using full rotations of the cube by 90 clockwise about a given face
        RotFront =pad(np.arange(48))# rotate full cube so that Left face becomes Up, Up becomes Right, Right becomes Down, Down becomes Left
        URDL = np.array([0,2,5,4])
        RotFront[URDL,:] = RotFront[np.roll(URDL,1),:]
        RotFront = unpad(RotFront)
        RotBack = np.argsort(RotFront)
        RotLeft = pad(np.arange(48))
        UFDB = np.array([0,1,5,3])
        RotLeft[UFDB,:] = RotLeft[np.roll(UFDB,1),:]
        RotLeft = unpad(RotLeft)
        RotRight = np.argsort(RotLeft)

        Fperm = RotRight[Uperm[RotLeft]]  # Fperm = RotLeft<-Uperm<-RotRight
        Rperm = RotBack[Uperm[RotFront]]  # Rperm = RotFront<-Uperm<-RotBack
        Bperm = RotLeft[Uperm[RotRight]]  # Bperm = RotRight<-Uperm<-RotLeft
        Lperm = RotFront[Uperm[RotBack]]  # Lperm = RotBack<-Uperm<-RotFront
        Dperm = RotRight[RotRight[Uperm[RotLeft[RotLeft]]]]  # Dperm = RotLeft<-RotLeft<-Uperm<-RotRight<-RotRight
        self.discrete_generators = [LazyPerm(perm) for perm in [Uperm,Fperm,Rperm,Bperm,Lperm,Dperm]]
        super().__init__()

@export
class ZksZnxZn(Group):
    """ One of the original GCNN groups ℤₖ⋉(ℤₙ×ℤₙ) for translation in x,y
        and rotation with the discrete 90 degree rotations (k=4) or 180 degree (k=2)"""
    def __init__(self,k,n):
        Zn = Z(n)
        Zk = Z(k)
        nshift = Zn.discrete_generators[0]
        kshift = Zk.discrete_generators[0]
        In = I(n)
        Ik = I(k)
        assert k in [2,4]
        self.discrete_generators = [LazyKron([Ik,nshift,In]),LazyKron([Ik,In,nshift]),LazyKron([kshift,Rot90(n,4//k)])]
        super().__init__(k,n)

@export
class Embed(Group):
    """ A method to embed a given base group representation in larger vector space.
    Inputs: 
    G: the group (and base representation) to embed
    d: the dimension in which to embed
    slice: a slice object specifying which dimensions G acts on."""
    def __init__(self,G,d,slice):
        self.lie_algebra = np.zeros((G.lie_algebra.shape[0],d,d))
        self.discrete_generators = np.zeros((G.discrete_generators.shape[0],d,d))
        self.discrete_generators += np.eye(d)
        self.lie_algebra[:,slice,slice] = G.lie_algebra
        self.discrete_generators[:,slice,slice]  =G.discrete_generators
        self.name = f"{G}_R{d}"
        super().__init__()
   
    def __repr__(self):
        return self.name

@export
def SO2eR3():
    """ SO(2) embedded in R^3 with rotations about z axis"""
    return Embed(SO(2),3,slice(2))

@export
def O2eR3():
    """ O(2) embedded in R^3 with rotations about z axis"""
    return Embed(O(2),3,slice(2))

@export
def DkeR3(k):
    """ Dihedral D(k) embedded in R^3 with rotations about z axis"""
    return Embed(D(k),3,slice(2))


class DirectProduct(Group):
    def __init__(self,G1,G2):
        I1,I2 = I(G1.d),I(G2.d)
        self.lie_algebra = [LazyKronsum([A1,0*I2]) for A1 in G1.lie_algebra]+[LazyKronsum([0*I1,A2]) for A2 in G2.lie_algebra]
        self.discrete_generators = [LazyKron([M1,I2]) for M1 in G1.discrete_generators]+[LazyKron([I1,M2]) for M2 in G2.discrete_generators]
        self.names = (repr(G1),repr(G2))
        super().__init__()

    def __repr__(self):
        return f"{self.names[0]}x{self.names[1]}"

class WreathProduct(Group):
    def __init__(self,G1,G2):
        raise NotImplementedError

class SemiDirectProduct(Group):
    def __init__(self,G1,G2,phi):
        raise NotImplementedError